#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

# %% [markdown]
# # Demonstration of a non-linear regression using NIFTy.re

# %%
import jax
import matplotlib.pyplot as plt
from jax import numpy as jnp
import numpy as np
from jax import random
from functools import partial
import operator as op

import nifty8.re as jft

jax.config.update("jax_enable_x64", True)

# %%
seed = 42
key = random.PRNGKey(seed)


class NonLinearRegression(jft.Model):
    def __init__(self, slope_mean, slope_std, intercept_min, intercept_max, x):
        self.slope = jft.LogNormalPrior(slope_mean, slope_std, name="slope")
        self.intercept = jft.UniformPrior(
            intercept_min, intercept_max, name="intercept"
        )
        self.x = x
        super().__init__(init=self.slope.init | self.intercept.init)

    def __call__(self, xi, *, x=None):
        x = x if x is not None else self.x
        return x * self.slope(xi) + self.intercept(xi)


key, sk = random.split(key)
x = random.uniform(sk, (50,), float, -4.0, 4.0)
nlr = NonLinearRegression(3.0, 2.0, -5.0, 5.0, x)

# %%
noise_std = 5

key_tr, key_n = random.split(random.PRNGKey(31415))
y = nlr(nlr.init(key_tr)) + noise_std * random.normal(key_n, x.shape, x.dtype)

# %%
fig, ax = plt.subplots()
ax.plot(x, y, color="dodgerblue", linestyle="None", marker=".", markersize=8)
ax.set_xlabel("x")
ax.set_ylabel("y")
fig.tight_layout()
fig.savefig("nonlinear_regression_data.png")
plt.show()

# %%
lh = jft.Gaussian(y, noise_std_inv=partial(op.mul, 1.0 / noise_std)).amend(nlr)

# %%
key, ki, ko = random.split(key, 3)

delta = 1e-4
samples_opt, st = jft.optimize_kl(
    lh,
    jft.Vector(lh.init(ki)),
    key=ko,
    n_total_iterations=5,
    n_samples=12,
    draw_linear_kwargs=dict(
        cg_kwargs=dict(absdelta=delta * jft.size(lh.domain) / 10.0),
    ),
    nonlinearly_update_kwargs=dict(
        minimize_kwargs=dict(xtol=delta, cg_kwargs=dict(name=None))
    ),
    kl_kwargs=dict(
        minimize_kwargs=dict(name="M", xtol=delta, cg_kwargs=dict(name=None))
    ),
    sample_mode="nonlinear_resample",
)

# %%
x_p = np.linspace(x.min(), x.max(), 500)
y_unnoised_samples = jax.vmap(partial(nlr, x=x_p))(samples_opt.samples)

fig, ax = plt.subplots()
ax.plot(x, y, color="dodgerblue", linestyle="None", marker=".", markersize=8)
ax.plot(x_p, y_unnoised_samples.mean(axis=0), color="black")
qs = (0.16, 0.84)
ax.fill_between(
    x_p, *np.quantile(y_unnoised_samples, qs, axis=0), color="gray", alpha=0.3
)
ax.set_xlabel("x")
ax.set_ylabel("y")
fig.tight_layout()
fig.savefig("nonlinear_regression_posterior.png")
plt.show()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import pytest

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize
space = list2fixture([
    ift.GLSpace(5),
    ift.RGSpace(3, distances=.789),
    ift.RGSpace([4, 4], distances=.789)
])
_h_RG_spaces = [
    ift.RGSpace(7, distances=0.2, harmonic=True),
    ift.RGSpace((3, 4), distances=(.2, .3), harmonic=True)
]
_h_spaces = _h_RG_spaces + [ift.LMSpace(17)]
space1 = space
seed = list2fixture([4, 78])
ntries = 10


def testBasics(space, seed):
    with ift.random.Context(seed):
        s = ift.from_random(space, 'normal')
        var = ift.Linearization.make_var(s)
        model = ift.ScalingOperator(var.target, 6.)
        ift.extra.check_operator(model, var.val, ntries=ntries)


@pmp('type1', ['Variable', 'Constant'])
@pmp('type2', ['Variable'])
def testBinary(type1, type2, space, seed):
    with ift.random.Context(seed):
        dom1 = ift.MultiDomain.make({'s1': space})
        dom2 = ift.MultiDomain.make({'s2': space})
        dom = ift.MultiDomain.union((dom1, dom2))
        select_s1 = ift.ducktape(None, dom1, "s1")
        select_s2 = ift.ducktape(None, dom2, "s2")
        model = select_s1*select_s2
        pos = ift.from_random(dom, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        model = select_s1 + select_s2
        pos = ift.from_random(dom, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        model = select_s1.scale(3.)
        pos = ift.from_random(dom1, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        model = ift.ScalingOperator(space, 2.456)(select_s1*select_s2)
        pos = ift.from_random(dom, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        model = (2.456*(select_s1*select_s2)).ptw("sigmoid")
        pos = ift.from_random(dom, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        pos = ift.from_random(dom, "normal")
        model = ift.OuterProduct(ift.makeDomain(space), pos['s1'])
        ift.extra.check_operator(model, pos['s2'], ntries=ntries)
        model = select_s1**2
        pos = ift.from_random(dom1, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        model = select_s1.clip(-1, 1)
        pos = ift.from_random(dom1, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        f = ift.from_random(space, "normal")
        model = select_s1.clip(f-0.1, f+1.)
        pos = ift.from_random(dom1, "normal")
        ift.extra.check_operator(model, pos, ntries=ntries)
        if isinstance(space, ift.RGSpace):
            model = ift.FFTOperator(space)(select_s1*select_s2)
            pos = ift.from_random(dom, "normal")
            ift.extra.check_operator(model, pos, ntries=ntries)


def testInverseGamma(space, seed):
    with ift.random.Context(seed):
        pos = ift.from_random(space, 'normal')
        alpha = 1.5
        q = 0.73
        model = ift.InverseGammaOperator(space, alpha, q)
        ift.extra.check_operator(model, pos, ntries=20)
        model = ift.LogInverseGammaOperator(space, alpha, q)
        ift.extra.check_operator(model, pos, ntries=20)
        model = ift.GammaOperator(space, alpha=alpha, theta=q)
        ift.extra.check_operator(model, pos, ntries=20)


@pmp("loc", [0, 13.2])
@pmp("scale", [1, 551.09])
@pmp("op", [ift.UniformOperator, ift.LaplaceOperator])
def testSpecialDistributionOps(space, seed, loc, scale, op):
    with ift.random.Context(seed):
        pos = ift.from_random(space, 'normal')
        model = op(space, loc, scale)
        ift.extra.check_operator(model, pos, ntries=20)


@pmp('neg', [True, False])
def testAdder(space, seed, neg):
    with ift.random.Context(seed):
        f = ift.from_random(space, 'normal')
        f1 = ift.from_random(space, 'normal')
        op = ift.Adder(f1, neg)
        ift.extra.check_operator(op, f, ntries=ntries)
        op = ift.Adder(f1.val.ravel()[0], neg=neg, domain=space)
        ift.extra.check_operator(op, f, ntries=ntries)


def testAbs(space, seed):
    with ift.random.Context(seed):
        f = ift.from_random(space, 'normal')
        op = abs(ift.ScalingOperator(space, 1.))
        ift.extra.check_operator(op, f, ntries=ntries)


@pmp('target', [ift.RGSpace(64, distances=.789, harmonic=True),
                ift.RGSpace([10, 10], distances=.789, harmonic=True)])
@pmp('causal', [True, False])
@pmp('minimum_phase', [True, False])
def testDynamicModel(target, causal, minimum_phase, seed):
    with ift.random.Context(seed):
        dct = {'target': target,
               'harmonic_padding': None,
               'sm_s0': 3.,
               'sm_x0': 1.,
               'key': 'f',
               'causal': causal,
               'minimum_phase': minimum_phase}
        model, _ = ift.dynamic_operator(**dct)
        pos = ift.from_random(model.domain, 'normal')
        ift.extra.check_operator(model, pos, ntries=ntries)
        if len(target.shape) > 1:
            dct = {
                'target': target,
                'harmonic_padding': None,
                'sm_s0': 3.,
                'sm_x0': 1.,
                'key': 'f',
                'lightcone_key': 'c',
                'sigc': 1.,
                'quant': 5,
                'causal': causal,
                'minimum_phase': minimum_phase
            }
            dct['lightcone_key'] = 'c'
            dct['sigc'] = 1.
            dct['quant'] = 5
            model, _ = ift.dynamic_lightcone_operator(**dct)
            pos = ift.from_random(model.domain, 'normal')
            ift.extra.check_operator(model, pos, ntries=ntries)


@pmp('h_space', _h_spaces)
@pmp('specialbinbounds', [True, False])
@pmp('logarithmic', [True, False])
@pmp('nbin', [3, None])
def testNormalization(h_space, specialbinbounds, logarithmic, nbin):
    if not specialbinbounds and (not logarithmic or nbin is not None):
        return
    binbounds = None
    if specialbinbounds:
        binbounds = ift.PowerSpace.useful_binbounds(h_space, logarithmic, nbin)
    dom = ift.PowerSpace(h_space, binbounds)
    op = ift.library.correlated_fields._Normalization(dom)
    pos = 0.1 * ift.from_random(op.domain, 'normal')
    ift.extra.check_operator(op, pos, ntries=10)


@pmp('N', [1, 20])
def testLognormalTransform(N):
    op = ift.LognormalTransform(1, 0.2, '', N)
    loc = ift.from_random(op.domain)
    ift.extra.check_operator(op, loc, ntries=10)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .line_search import LineSearch
from .minimizer import Minimizer


class NonlinearCG(Minimizer):
    """Nonlinear Conjugate Gradient scheme according to Polak-Ribiere.

    Algorithm 5.4 from Nocedal & Wright.

    Parameters
    ----------
    controller : IterationController
        Object that decides when to terminate the minimization.
    beta_heuristics : str
        One of 'Polak-Ribiere', 'Fletcher-Reeves', 'Hestenes-Stiefel' or '5.49'

    Notes
    -----
    No restarting procedure has been implemented yet.

    References
    ----------
    Jorge Nocedal & Stephen Wright, "Numerical Optimization", Second Edition,
    2006, Springer-Verlag New York
    """

    def __init__(self, controller, beta_heuristics='Polak-Ribiere'):
        valid_beta_heuristics = ['Polak-Ribiere', 'Fletcher-Reeves',
                                 'Hestenes-Stiefel', "5.49"]
        if not (beta_heuristics in valid_beta_heuristics):
            raise ValueError("beta heuristics must be either 'Polak-Ribiere', "
                             "'Fletcher-Reeves', 'Hestenes-Stiefel, or '5.49'")
        self._beta_heuristic = beta_heuristics
        self._controller = controller
        self._line_searcher = LineSearch(c2=0.1)

    def __call__(self, energy):
        controller = self._controller
        status = controller.start(energy)
        if status != controller.CONTINUE:
            return energy, status
        f_k_minus_1 = None

        p = -energy.gradient

        while True:
            grad_old = energy.gradient
            f_k = energy.value
            energy, success = self._line_searcher.perform_line_search(
                energy, p, f_k_minus_1)
            if not success:
                return energy, controller.ERROR
            f_k_minus_1 = f_k
            status = self._controller.check(energy)
            if status != controller.CONTINUE:
                return energy, status
            grad_new = energy.gradient

            if self._beta_heuristic == 'Hestenes-Stiefel':
                # Eq. (5.46) in Nocedal & Wright.
                beta = max(0.0, (grad_new.s_vdot(grad_new-grad_old) /
                                 (grad_new-grad_old).s_vdot(p)).real)
            elif self._beta_heuristic == 'Polak-Ribiere':
                # Eq. (5.44) in Nocedal & Wright. (with (5.45) additionally)
                beta = max(0.0, (grad_new.s_vdot(grad_new-grad_old) /
                                 (grad_old.s_vdot(grad_old))).real)
            elif self._beta_heuristic == 'Fletcher-Reeves':
                # Eq. (5.41a) in Nocedal & Wright.
                beta = (grad_new.s_vdot(grad_new)/(grad_old.s_vdot(grad_old))).real
            else:
                # Eq. (5.49) in Nocedal & Wright.
                beta = (grad_new.s_vdot(grad_new) /
                        ((grad_new-grad_old).s_vdot(p))).real
            p = beta*p - grad_new



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..field import Field
from ..logger import logger
from ..multi_field import MultiField
from ..utilities import iscomplextype
from .iteration_controllers import IterationController
from .minimizer import Minimizer


def _multiToArray(fld):
    szall = sum(2*v.size if iscomplextype(v.dtype) else v.size
                for v in fld.values())
    res = np.empty(szall, dtype=np.float64)
    ofs = 0
    for val in fld.values():
        sz2 = 2*val.size if iscomplextype(val.dtype) else val.size
        locdat = val.val.reshape(-1)
        if iscomplextype(val.dtype):
            locdat = locdat.view(locdat.real.dtype)
        res[ofs:ofs+sz2] = locdat
        ofs += sz2
    return res


def _toArray(fld):
    if isinstance(fld, Field):
        return fld.val.reshape(-1)
    return _multiToArray(fld)


def _toArray_rw(fld):
    if isinstance(fld, Field):
        return fld.val_rw().reshape(-1)
    return _multiToArray(fld)


def _toField(arr, template):
    if isinstance(template, Field):
        return Field(template.domain, arr.reshape(template.shape).copy())
    ofs = 0
    res = []
    for v in template.values():
        sz2 = 2*v.size if iscomplextype(v.dtype) else v.size
        locdat = arr[ofs:ofs+sz2].copy()
        if iscomplextype(v.dtype):
            locdat = locdat.view(np.complex128)
        res.append(Field(v.domain, locdat.reshape(v.shape)))
        ofs += sz2
    return MultiField(template.domain, tuple(res))


class _MinHelper:
    def __init__(self, energy):
        self._energy = energy
        self._domain = energy.position.domain

    def _update(self, x):
        pos = _toField(x, self._energy.position)
        if (pos != self._energy.position).s_any():
            self._energy = self._energy.at(pos)

    def fun(self, x):
        self._update(x)
        return self._energy.value

    def jac(self, x):
        self._update(x)
        return _toArray_rw(self._energy.gradient)

    def hessp(self, x, p):
        self._update(x)
        res = self._energy.apply_metric(_toField(p, self._energy.position))
        return _toArray_rw(res)


class _ScipyMinimizer(Minimizer):
    """Scipy-based minimizer

    Parameters
    ----------
    method     : str
        The selected Scipy minimization method.
    options    : dictionary
        A set of custom options for the selected minimizer.
    """

    def __init__(self, method, options, need_hessp, bounds):
        self._method = method
        self._options = options
        self._need_hessp = need_hessp
        self._bounds = bounds

    def __call__(self, energy):
        import scipy.optimize as opt
        hlp = _MinHelper(energy)
        energy = None  # drop handle, since we don't need it any more
        bounds = None
        if self._bounds is not None:
            if len(self._bounds) == 2:
                lo = self._bounds[0]
                hi = self._bounds[1]
                bounds = [(lo, hi)]*hlp._energy.position.size
            else:
                raise ValueError("unrecognized bounds")

        x = _toArray_rw(hlp._energy.position)
        hessp = hlp.hessp if self._need_hessp else None
        r = opt.minimize(hlp.fun, x, method=self._method, jac=hlp.jac,
                         hessp=hessp, options=self._options, bounds=bounds)
        if not r.success:
            logger.error("Problem in Scipy minimization: {}".format(r.message))
            return hlp._energy, IterationController.ERROR
        return hlp._energy, IterationController.CONVERGED


def L_BFGS_B(ftol, gtol, maxiter, maxcor=10, disp=False, bounds=None):
    """Returns a _ScipyMinimizer object carrying out the L-BFGS-B algorithm.

    See Also
    --------
    _ScipyMinimizer
    """
    options = {"ftol": ftol, "gtol": gtol, "maxiter": maxiter,
               "maxcor": maxcor, "disp": disp}
    return _ScipyMinimizer("L-BFGS-B", options, False, bounds)


class _ScipyCG(Minimizer):
    """Returns a _ScipyMinimizer object carrying out the conjugate gradient
    algorithm as implemented by SciPy.

    This class is only intended for double-checking NIFTy's own conjugate
    gradient implementation and should not be used otherwise.
    """
    def __init__(self, tol, maxiter):
        self._tol = tol
        self._maxiter = maxiter

    def __call__(self, energy, preconditioner=None):
        from scipy.sparse.linalg import LinearOperator as scipy_linop
        from scipy.sparse.linalg import cg

        from .quadratic_energy import QuadraticEnergy
        if not isinstance(energy, QuadraticEnergy):
            raise ValueError("need a quadratic energy for CG")

        class mymatvec:
            def __init__(self, op):
                self._op = op

            def __call__(self, inp):
                return _toArray(self._op(_toField(inp, energy.position)))

        op = energy._A
        b = _toArray(energy._b)
        sx = _toArray(energy.position)
        sci_op = scipy_linop(shape=(op.domain.size, op.target.size),
                             matvec=mymatvec(op))
        prec_op = None
        if preconditioner is not None:
            prec_op = scipy_linop(shape=(op.domain.size, op.target.size),
                                  matvec=mymatvec(preconditioner))
        res, stat = cg(sci_op, b, x0=sx, rtol=self._tol, M=prec_op,
                       maxiter=self._maxiter, atol=0.)
        stat = (IterationController.CONVERGED if stat >= 0 else
                IterationController.ERROR)
        return energy.at(_toField(res, energy.position)), stat



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Max-Planck-Society
# Author: Matteo Guardiani

import nifty8 as ift
import numpy as np
import pytest

pmp = pytest.mark.parametrize
from nifty8.operator_spectrum import _DomRemover


def _explicify(operator):
    tmp = _DomRemover(operator.domain)
    operator = operator @ tmp.adjoint
    tmp = _DomRemover(operator.target)
    operator = tmp @ operator
    identity = np.identity(operator.domain.size, dtype=np.float64)
    res = []
    for v in identity:
        res.append(operator(ift.makeField(operator.domain, v)).val)
    return np.vstack(res).T


class LinearResponse(ift.LinearOperator):
    """Calculates values of a polynomial parameterized by input at sampling
    points.

    Parameters
    ----------
    domain: UnstructuredDomain
        The domain on which the coefficients of the polynomial are defined.
    sampling_points: Numpy array
        x-values of the sampling points.
    """

    def __init__(self, intercept, slope, sampling_points):
        # FIXME: Add input checks
        self.intercept_key = intercept.domain.keys()[0]
        self.slope_key = slope.domain.keys()[0]
        domain = {self.intercept_key: intercept.domain[self.intercept_key],
                  self.slope_key: slope.domain[self.slope_key]}
        self.sampling_points = sampling_points

        self._domain = ift.MultiDomain.make(domain)
        tgt = ift.RGSpace(self.sampling_points.shape)
        self._target = ift.DomainTuple.make(tgt)
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)

        if mode == self.TIMES:
            intercept = x.val[self.intercept_key]
            slope = x.val[self.slope_key]
            return ift.makeField(self._tgt(mode), intercept + slope * self.sampling_points)

        res = np.vstack((np.ones(self.sampling_points.shape[0]), self.sampling_points)).dot(x.val)
        return ift.makeField(self._tgt(mode), {'intercept': res[0], 'slope': res[1]})


def test_estimate_evidence_lower_bound():
    # Set up signal
    n_datapoints = 8
    data_space = ift.RGSpace((n_datapoints,))

    q = -1.
    m_slope = 1.5

    sigma_q = 1.5
    sigma_m = 1.8
    intercept = ift.NormalTransform(0., sigma_q, "intercept").ducktape_left("intercept")
    slope = ift.NormalTransform(0., sigma_m, "slope").ducktape_left("slope")

    x = ift.random.current_rng().random(n_datapoints) * 10
    y = q + m_slope * x
    linear_response = LinearResponse(intercept, slope, x)
    # ift.extra.check_linear_operator(linear_response)

    linear_response_on_signal = linear_response @ (intercept + slope)  # In general not a linear operator

    R = _explicify(linear_response_on_signal)
    d = ift.makeField(data_space, y)
    noise_level = 0.8
    N = ift.makeOp(ift.makeField(data_space, noise_level ** 2), sampling_dtype=float)
    noise = N.draw_sample()
    data = d + noise

    N_inv = _explicify(N.inverse)

    S = np.identity(2)
    S_inv = np.identity(2)

    D_inv = R.T @ N_inv @ R + S_inv
    D = np.linalg.inv(D_inv)

    j = R.T @ (N_inv @ data.val)
    m = D @ j
    m_dag_j = np.dot(m, j)

    det_2pi_D = np.linalg.det(2 * np.pi * D)
    det_2pi_S = np.linalg.det(2 * np.pi * S)

    H_0 = 0.5 * (data.s_vdot(N.inverse(data)) + n_datapoints * np.log(2 * np.pi * noise_level ** 2) + np.log(
        det_2pi_S) - m_dag_j)

    evidence = -H_0 + 0.5 * np.log(det_2pi_D)
    nifty_adjusted_evidence = evidence + 0.5 * n_datapoints * np.log(2 * np.pi * noise_level ** 2)
    likelihood_energy = ift.GaussianEnergy(data=data, inverse_covariance=N.inverse) @ linear_response_on_signal

    # Minimization parameters
    ic_sampling = ift.AbsDeltaEnergyController(name="Sampling (linear)", deltaE=1e-8, iteration_limit=2)
    ic_newton = ift.AbsDeltaEnergyController(name='Newton', deltaE=1e-5, convergence_level=2, iteration_limit=100)

    minimizer = ift.NewtonCG(ic_newton)
    minimizer_sampling = None

    n_iterations = 2
    n_samples = 2

    samples = ift.optimize_kl(likelihood_energy, n_iterations, n_samples, minimizer, ic_sampling, minimizer_sampling)

    # Estimate the ELBO
    elbo, stats = ift.estimate_evidence_lower_bound(ift.StandardHamiltonian(lh=likelihood_energy), samples, 2)
    assert (stats['elbo_lw'].val <= nifty_adjusted_evidence <= stats['elbo_up'].val)



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import dataclasses
import operator
from functools import partial
from typing import Any, Callable, Optional, Tuple, Union

from jax import numpy as jnp
from jax.tree_util import Partial, tree_map

from .likelihood import Likelihood
from .logger import logger
from .model import LazyModel
from .tree_math import ShapeWithDtype, result_type, sum, vdot


def _standard_t(nwr, dof):
    res = (nwr.conj() * nwr).real / dof
    return sum(tree_map(jnp.log1p, res) * (dof + 1)) / 2


def _shape_w_fixed_dtype(dtype):
    def shp_w_dtp(e):
        return ShapeWithDtype(jnp.shape(e), dtype)

    return shp_w_dtp


@Partial
def _identity(x):
    return x


def _get_cov_inv_and_std_inv(
    cov_inv: Optional[Callable],
    std_inv: Optional[Callable],
    primals=None
) -> Tuple[Union[Partial, LazyModel], Union[Partial, LazyModel]]:
    if cov_inv is None and std_inv is None:
        _cov_inv, _std_inv = _identity, _identity

    if not callable(cov_inv) and cov_inv is not None:
        msg = "assuming the specified inverse covariance is diagonal"
        logger.warning(msg)
        _cov_inv = Partial(operator.mul, cov_inv)
    elif cov_inv is None:
        wm = (
            "assuming a diagonal covariance matrix"
            ";\nsetting `cov_inv` to `std_inv(ones_like(data))**2`"
        )
        logger.warning(wm)
        # Note, `_std_inv` is not properly initialized yet
        si = std_inv if std_inv is not None else _std_inv
        noise_std_inv_sq = si(
            tree_map(jnp.real, tree_map(jnp.ones_like, primals))
        )**2
        _cov_inv = Partial(operator.mul, noise_std_inv_sq)
    else:
        _cov_inv = cov_inv if isinstance(cov_inv,
                                         (Partial,
                                          LazyModel)) else Partial(cov_inv)

    if not callable(std_inv) and std_inv is not None:
        msg = "assuming the specified sqrt of the inverse covariance is diagonal"
        logger.warning(msg)
        _std_inv = Partial(operator.mul, std_inv)
    elif std_inv is None:
        wm = (
            "assuming a diagonal covariance matrix"
            ";\nsetting `std_inv` to `cov_inv(ones_like(data))**0.5`"
        )
        logger.warning(wm)
        noise_cov_inv_sqrt = tree_map(
            jnp.sqrt,
            _cov_inv(tree_map(jnp.real, tree_map(jnp.ones_like, primals)))
        )
        _std_inv = Partial(operator.mul, noise_cov_inv_sqrt)
    else:
        _std_inv = std_inv if isinstance(std_inv,
                                         (Partial,
                                          LazyModel)) else Partial(std_inv)

    assert all(
        isinstance(c, (Partial, LazyModel)) for c in (_cov_inv, _std_inv)
    )
    return _cov_inv, _std_inv


class Gaussian(Likelihood):
    """Gaussian likelihood of the data

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.
    noise_cov_inv : callable acting on type of data
        Function applying the inverse noise covariance of the Gaussian.
    noise_std_inv : callable acting on type of data
        Function applying the square root of the inverse noise covariance.

    Notes
    -----
    If `noise_std_inv` is `None` it is inferred by assuming a diagonal noise
    covariance, i.e. by applying it to a vector of ones and taking the square
    root. If both `noise_cov_inv` and `noise_std_inv` are `None`, a unit
    covariance is assumed.

    See :class:`Likelihood` for details on the properties.
    """
    data: Any = dataclasses.field(metadata=dict(static=False))
    noise_cov_inv: Callable = dataclasses.field(metadata=dict(static=False))
    noise_std_inv: Callable = dataclasses.field(metadata=dict(static=False))

    def __init__(
        self,
        data,
        noise_cov_inv: Optional[Callable] = None,
        noise_std_inv: Optional[Callable] = None
    ):
        self.data = data
        noise_cov_inv, noise_std_inv = _get_cov_inv_and_std_inv(
            noise_cov_inv, noise_std_inv, data
        )
        self.noise_cov_inv = noise_cov_inv
        self.noise_std_inv = noise_std_inv
        shp = tree_map(ShapeWithDtype.from_leave, data)
        super().__init__(domain=shp, lsm_tangents_shape=shp)

    def energy(self, primals):
        p_res = self.data - primals
        return 0.5 * vdot(p_res, self.noise_cov_inv(p_res)).real

    def normalized_residual(self, primals):
        return self.noise_std_inv(self.data - primals)

    def metric(self, primals, tangents):
        return self.noise_cov_inv(tangents)

    def left_sqrt_metric(self, primals, tangents):
        return self.noise_std_inv(tangents)

    def transformation(self, primals):
        return self.noise_std_inv(primals)


class StudentT(Likelihood):
    """Student's t likelihood of the data

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.
    dof : tree-like structure of jnp.ndarray and float
        Degree-of-freedom parameter of Student's t distribution.
    noise_cov_inv : callable acting on type of data
        Function applying the inverse noise covariance of the Gaussian.
    noise_std_inv : callable acting on type of data
        Function applying the square root of the inverse noise covariance.

    Notes
    -----
    If `noise_std_inv` is `None` it is inferred by assuming a diagonal noise
    covariance, i.e. by applying it to a vector of ones and taking the square
    root. If both `noise_cov_inv` and `noise_std_inv` are `None`, a unit
    covariance is assumed.

    See :class:`Likelihood` for details on the properties.
    """
    data: Any = dataclasses.field(metadata=dict(static=False))
    dof: Any = dataclasses.field(metadata=dict(static=False))
    noise_cov_inv: Callable = dataclasses.field(metadata=dict(static=False))
    noise_std_inv: Callable = dataclasses.field(metadata=dict(static=False))

    def __init__(
        self,
        data,
        dof,
        noise_cov_inv: Optional[Callable] = None,
        noise_std_inv: Optional[Callable] = None,
    ):
        self.data = data
        self.dof = dof
        noise_cov_inv, noise_std_inv = _get_cov_inv_and_std_inv(
            noise_cov_inv, noise_std_inv, data
        )
        self.noise_cov_inv = noise_cov_inv
        self.noise_std_inv = noise_std_inv
        shp = tree_map(ShapeWithDtype.from_leave, data)
        super().__init__(domain=shp, lsm_tangents_shape=shp)

    def energy(self, primals):
        return _standard_t(self.noise_std_inv(self.data - primals), self.dof)

    def metric(self, primals, tangents):
        return self.noise_cov_inv((self.dof + 1) / (self.dof + 3) * tangents)

    def left_sqrt_metric(self, primals, tangents):
        return self.noise_std_inv(
            ((self.dof + 1) / (self.dof + 3))**0.5 * tangents
        )

    def normalized_residual(self, primals):
        return self.left_sqrt_metric(None, self.data - primals)

    def transformation(self, primals):
        return self.noise_std_inv(
            ((self.dof + 1) / (self.dof + 3))**0.5 * primals
        )


class Poissonian(Likelihood):
    """Computes the negative log-likelihood, i.e. the Hamiltonians of an
    expected count Vector constrained by Poissonian count data.

    Represents up to an f-independent term :math:`log(d!)`:

    .. math ::
        E(f) = -\\log \\text{Poisson}(d|f) = \\sum f - d^\\dagger \\log(f),

    where f is a Vector in data space of the expectation values for the counts.

    Parameters
    ----------
    data : jnp.ndarray or tree-like structure of jnp.ndarray and float
        Data Vector with counts. Needs to have integer dtype and all values need
        to be non-negative.
    sampling_dtype : dtype, optional
        Data-type for sampling.


    See :class:`Likelihood` for details on the properties.
    """
    data: Any = dataclasses.field(metadata=dict(static=False))

    def __init__(self, data, sampling_dtype=float):
        dtp = result_type(data)
        if not jnp.issubdtype(dtp, jnp.integer):
            raise TypeError("`data` of invalid type")
        if sum(tree_map(lambda x: jnp.any(x < 0), data)):
            raise ValueError("`data` must not be negative")
        self.data = data
        shp = tree_map(_shape_w_fixed_dtype(sampling_dtype), data)
        super().__init__(domain=shp, lsm_tangents_shape=shp)

    def energy(self, primals):
        return sum(primals) - vdot(tree_map(jnp.log, primals), self.data)

    def metric(self, primals, tangents):
        return tangents / primals

    def left_sqrt_metric(self, primals, tangents):
        return tangents / primals**0.5

    def normalized_residual(self, primals):
        return self.left_sqrt_metric(primals, self.data - primals)

    def transformation(self, primals):
        return 2. * primals**0.5


class VariableCovarianceGaussian(Likelihood):
    """Gaussian likelihood of the data with a variable covariance

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.
    iscomplex: Boolean, optional
        Whether the parameters are complex-valued.

    Notes
    -----
    **The likelihood acts on a tuple of (mean, std_inv)**.

    See :class:`Likelihood` for details on the properties.
    """
    data: Any = dataclasses.field(metadata=dict(static=False))
    iscomplex: bool = False

    def __init__(self, data, iscomplex=False):
        # TODO: make configurable whether `std_inv` or `std` is passed
        self.data = data
        self.iscomplex = iscomplex
        shp = tree_map(ShapeWithDtype.from_leave, (data, data.real))
        super().__init__(domain=shp, lsm_tangents_shape=shp)

    def energy(self, primals):
        res = (self.data - primals[0]) * primals[1]
        fct = 1 + self.iscomplex
        return 0.5 * vdot(res,
                          res).real - fct * sum(tree_map(jnp.log, primals[1]))

    def metric(self, primals, tangents):
        fct = 2 * (1 + self.iscomplex)
        prim_std_inv_sq = primals[1]**2
        res = (
            prim_std_inv_sq * tangents[0], fct * tangents[1] / prim_std_inv_sq
        )
        return type(primals)(res)

    def left_sqrt_metric(self, primals, tangents):
        fct = jnp.sqrt(2)**(1 + self.iscomplex)
        res = (primals[1] * tangents[0], fct * tangents[1] / primals[1])
        return type(primals)(res)

    def transformation(self, primals):
        """
        Notes
        -----
        A global transformation to Euclidean space does not exist. A local
        approximation invoking the residual is used instead.
        """
        # TODO: test by drawing synthetic data that actually follows the
        # noise-cov and then average over it
        fct = 1 + self.iscomplex
        res = (
            primals[1] * (primals[0] - self.data),
            fct * tree_map(jnp.log, primals[1])
        )
        return type(primals)(res)

    def normalized_residual(self, primals):
        return (self.data - primals[0]) * primals[1]


class VariableCovarianceStudentT(Likelihood):
    """Student's t likelihood of the data with a variable covariance

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and float
        Data with additive noise following a Gaussian distribution.
    dof : tree-like structure of jnp.ndarray and float
        Degree-of-freedom parameter of Student's t distribution.

    Notes
    -----
    **The likelihood acts on a tuple of (mean, std)**.

    See :class:`Likelihood` for details on the properties.
    """
    data: Any = dataclasses.field(metadata=dict(static=False))
    dof: Any = dataclasses.field(metadata=dict(static=False))

    def __init__(self, data, dof):
        # TODO: make configurable whether `std_inv` or `std` is passed
        self.data = data
        self.dof = dof
        shp = tree_map(ShapeWithDtype.from_leave, (data, data))
        super().__init__(domain=shp, lsm_tangents_shape=shp)

    def energy(self, primals):
        t = _standard_t((self.data - primals[0]) / primals[1], self.dof)
        t += sum(tree_map(jnp.log, primals[1]))
        return t

    def metric(self, primals, tangent):
        res = (
            tangent[0] * (self.dof + 1) / (self.dof + 3) / primals[1]**2,
            tangent[1] * 2 * self.dof / (self.dof + 3) / primals[1]**2
        )
        return type(primals)(res)

    def left_sqrt_metric(self, primals, tangents):
        cov = (
            (self.dof + 1) / (self.dof + 3) / primals[1]**2,
            2 * self.dof / (self.dof + 3) / primals[1]**2
        )
        res = (cov[0]**0.5 * tangents[0], cov[1]**0.5 * tangents[1])
        return type(primals)(res)

    def normalized_residual(self, primals):
        return (self.data - primals[0]
               ) / primals[1] * ((self.dof + 1) / (self.dof + 3))**0.5


class Categorical(Likelihood):
    """Categorical likelihood of the data, equivalent to cross entropy

    Parameters
    ----------
    data : tree-like structure of jnp.ndarray and int
        Which of the categories is the realized in the data. Must agree with the
        input shape except for the shape[axis] of the leafs
    axis : int
        Leaf-axis over which the categories are formed
    sampling_dtype : dtype, optional
        Data-type for sampling.


    See :class:`Likelihood` for details on the properties.
    """
    data: Any = dataclasses.field(metadata=dict(static=False))
    axis: int = -1

    def __init__(self, data, axis=-1, sampling_dtype=float):
        self.data = data
        self.axis = axis
        shp = tree_map(_shape_w_fixed_dtype(sampling_dtype), data)
        super().__init__(domain=shp, lsm_tangents_shape=shp)

    def energy(self, primals):
        from jax.nn import log_softmax

        def eval(p, d):
            logits = log_softmax(p, axis=self.axis)
            return -jnp.sum(jnp.take_along_axis(logits, d, self.axis))

        return sum(tree_map(eval, primals, self.data))

    def metric(self, primals, tangents):
        from jax.nn import softmax

        preds = tree_map(partial(softmax, axis=self.axis), primals)
        norm_term = tree_map(
            partial(jnp.sum, axis=self.axis, keepdims=True), preds * tangents
        )
        return preds * tangents - preds * sum(norm_term)

    def left_sqrt_metric(self, primals, tangents):
        from jax.nn import softmax

        # FIXME: not sure if this is really the square root
        sqrtp = tree_map(partial(softmax, axis=self.axis), primals)**0.5
        norm_term = tree_map(
            partial(jnp.sum, axis=self.axis, keepdims=True), sqrtp * tangents
        )
        norm_term = sum(norm_term)
        return sqrtp * (tangents - sqrtp * norm_term)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
# Authors: Philipp Frank
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..multi_domain import MultiDomain
from ..multi_field import MultiField
from .endomorphic_operator import EndomorphicOperator


class PartialConjugate(EndomorphicOperator):
    """Perform partial conjugation of a :class:`MultiField`

    Parameters
    ----------
    domain : MultiDomain
        The operator's input domain and output target
    conjugation_keys : iterable of string
        The keys of the :class:`MultiField` for which complex conjugation
        should be performed.
    """
    def __init__(self, domain, conjugation_keys):
        if not isinstance(domain, MultiDomain):
            raise ValueError("MultiDomain expected!")
        indom = (key in domain.keys() for key in conjugation_keys)
        if sum(indom) != len(conjugation_keys):
            raise ValueError("conjugation_keys not in domain!")
        self._domain = domain
        self._conjugation_keys = conjugation_keys
        self._capabilities = self._all_ops

    def apply(self, x, mode):
        self._check_input(x, mode)
        x = x.to_dict()
        for k in self._conjugation_keys:
            x[k] = x[k].conjugate()
        return MultiField.from_dict(x, self._domain)



import pytest

pytest.importorskip("jax")

from functools import partial

from jax import numpy as jnp
from jax.tree_util import tree_leaves
from numpy.testing import assert_array_equal

import nifty8.re as jft


def test_hmc_pytree():
    """Test sapmler output against known hash from previous commits."""
    initial_position = jnp.array([0.31415, 2.71828])

    sampler_init = partial(
        jft.HMCChain,
        potential_energy=lambda x: jft.vdot(x, x),
        inverse_mass_matrix=1.,
        step_size=0.193,
        num_steps=100
    )

    initial_position_py = jft.Vector(({"lvl0": initial_position}, ))
    smpl_w_pytree = sampler_init(position_proto=initial_position_py
                                ).generate_n_samples(
                                    key=321,
                                    initial_position=initial_position_py,
                                    num_samples=1000
                                )
    smpl_wo_pytree = sampler_init(position_proto=initial_position
                                 ).generate_n_samples(
                                     key=321,
                                     initial_position=initial_position,
                                     num_samples=1000
                                 )

    ts_w, ts_wo = tree_leaves(smpl_w_pytree), tree_leaves(smpl_wo_pytree)
    assert len(ts_w) == len(ts_wo)
    for w, wo in zip(ts_w, ts_wo):
        assert_array_equal(w, wo)


def test_nuts_pytree():
    """Test sapmler output against known hash from previous commits."""
    initial_position = jnp.array([0.31415, 2.71828])

    sampler_init = partial(
        jft.NUTSChain,
        potential_energy=lambda x: jft.vdot(x, x),
        inverse_mass_matrix=1.,
        step_size=0.193,
        max_tree_depth=10,
    )

    initial_position_py = jft.Vector(({"lvl0": initial_position}, ))
    smpl_w_pytree = sampler_init(position_proto=initial_position_py
                                ).generate_n_samples(
                                    key=323,
                                    initial_position=initial_position_py,
                                    num_samples=1000
                                )
    smpl_wo_pytree = sampler_init(position_proto=initial_position
                                 ).generate_n_samples(
                                     key=323,
                                     initial_position=initial_position,
                                     num_samples=1000
                                 )

    ts_w, ts_wo = tree_leaves(smpl_w_pytree), tree_leaves(smpl_wo_pytree)
    assert len(ts_w) == len(ts_wo)
    for w, wo in zip(ts_w, ts_wo):
        assert_array_equal(w, wo)


if __name__ == "__main__":
    test_hmc_pytree()
    test_nuts_pytree()



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from functools import partial

import pytest

pytest.importorskip("jax")

import jax
import numpy as np

import nifty8.re as jft
from nifty8.re.refine import util

pmp = pytest.mark.parametrize


@pmp("shape0", ((16, ), (13, 15), (11, 12, 13)))
@pmp("depth", (1, 2))
@pmp("_coarse_size", (3, 5, 7))
@pmp("_fine_size", (2, 4, 6))
@pmp("_fine_strategy", ("jump", "extend"))
def test_shape_translations(
    shape0, depth, _coarse_size, _fine_size, _fine_strategy
):
    kwargs = {
        "_coarse_size": _coarse_size,
        "_fine_size": _fine_size,
        "_fine_strategy": _fine_strategy
    }

    def cf(shape0, xi):
        chart = jft.CoordinateChart(
            shape0=shape0,
            depth=depth,
            distances0=(1., ) * len(shape0),
            **kwargs
        )
        return jft.RefinementField.apply(xi, chart=chart, kernel=lambda x: x)

    dom = util.get_refinement_shapewithdtype(shape0, depth, **kwargs)
    tgt = jax.eval_shape(partial(cf, shape0), dom)
    tgt_pred_shp = util.coarse2fine_shape(shape0, depth, **kwargs)
    assert tgt_pred_shp == tgt.shape
    assert dom[-1].size == tgt.size == np.prod(tgt_pred_shp)

    shape0_pred = util.fine2coarse_shape(tgt.shape, depth, **kwargs)
    dom_pred = util.get_refinement_shapewithdtype(shape0_pred, depth, **kwargs)
    tgt_pred = jax.eval_shape(partial(cf, shape0_pred), dom_pred)

    assert tgt.shape == tgt_pred.shape
    if _fine_strategy == "jump":
        assert shape0_pred == shape0
    else:
        assert _fine_strategy == "extend"
        assert all(s0_p <= s0 for s0_p, s0 in zip(shape0_pred, shape0))


@pmp("seed", (42, 45))
def test_gauss_kl(seed, n_resamples=100):
    rng = np.random.default_rng(seed)
    for _ in range(n_resamples):
        d = max(rng.poisson(4), 1)
        m_t = rng.normal(size=(d, d))
        m_t = m_t @ m_t.T
        scl = rng.lognormal(2., 3.)

        np.testing.assert_allclose(util.gauss_kl(m_t, m_t), 0., atol=1e-11)
        kl_rhs_scl = 0.5 * d * (np.log(scl) + 1. / scl - 1.)
        np.testing.assert_allclose(
            kl_rhs_scl, util.gauss_kl(m_t, scl * m_t), rtol=1e-10
        )
        kl_lhs_scl = 0.5 * d * (-np.log(scl) + scl - 1.)
        np.testing.assert_allclose(
            kl_lhs_scl, util.gauss_kl(scl * m_t, m_t), rtol=1e-10
        )



import pytest
import nifty8.re as jft
import numpy as np

pmp = pytest.mark.parametrize


@pmp(
    "params",
    (
        ({"t1": 2, 5: "foo"}, ("t1",), {"t1": 2}, {5: "foo"}),
        ({}, (), {}, {}),
        ({}, (42,), {}, {}),
        ({"t1": 2, 5: "foo"}, (), {}, {"t1": 2, 5: "foo"}),
        ({"t1": 2, 5: "foo"}, ("key",), {}, {"t1": 2, 5: "foo"}),
        ({"t1": 2, 5: "foo"}, ("t1", 5, False), {"t1": 2, 5: "foo"}, {}),
    ),
)
def test_dictsplit(params):
    dct, keys, sel, rest = params
    sel2, rest2 = jft.misc.split(dct, keys)
    assert sel2 == sel and rest2 == rest


@pmp(
    "params",
    (
        (np.zeros(10), True),
        (np.zeros((1, 1)), False),
        (np.zeros(()), False),
        ([2, 3, 4], True),
        (False, False),
        ((), True),
    ),
)
def test_is1d(params):
    obj, res = params
    assert jft.misc.is_iterable_of_non_iterables(obj) is res



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np
import pytest
from nifty8 import HPSpace, myassert
from numpy.testing import assert_almost_equal, assert_equal, assert_raises

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize
# [nside, expected]
CONSTRUCTOR_CONFIGS = [[
    2, {
        'nside': 2,
        'harmonic': False,
        'shape': (48,),
        'size': 48,
    }
], [5, {
    'nside': 5,
    'harmonic': False,
    'shape': (300,),
    'size': 300,
}], [1, {
    'nside': 1,
    'harmonic': False,
    'shape': (12,),
    'size': 12,
}], [0, {
    'error': ValueError
}]]


def test_property_ret_type():
    x = HPSpace(2)
    myassert(isinstance(getattr(x, 'nside'), int))


@pmp('nside, expected', CONSTRUCTOR_CONFIGS)
def test_constructor(nside, expected):
    if 'error' in expected:
        with assert_raises(expected['error']):
            HPSpace(nside)
    else:
        h = HPSpace(nside)
        for key, value in expected.items():
            assert_equal(getattr(h, key), value)


def test_dvol():
    assert_almost_equal(HPSpace(2).dvol, np.pi/12)



#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

# %%
from functools import partial

import jax.numpy as jnp
import matplotlib.pyplot as plt

import nifty8.re as jft


def loggaussian(x, mu, sigma):
    return -0.5 * (x - mu)**2 / sigma


def sum_of_gaussians(x, separation, sigma1, sigma2):
    return -jnp.logaddexp(
        loggaussian(x, 0, sigma1), loggaussian(x, separation, sigma2)
    )


ham = partial(sum_of_gaussians, separation=10., sigma1=1., sigma2=1.)

N = 100000
SEED = 43
EPS = 0.3

subplots = (2, 2)
fig_width_pt = 426  # pt (a4paper, and such)
# fig_width_pt = 360 # pt
inches_per_pt = 1 / 72.27
fig_width_in = 0.9 * fig_width_pt * inches_per_pt
fig_height_in = fig_width_in * 0.618 * (subplots[0] / subplots[1])
fig_dims = (fig_width_in, fig_height_in * 1.5)

fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
    subplots[0],
    subplots[1],
    sharex='col',
    figsize=fig_dims,
    gridspec_kw={'width_ratios': [1, 2]}
)

# %%
nuts_sampler = jft.NUTSChain(
    potential_energy=ham,
    inverse_mass_matrix=5.,
    position_proto=jnp.array(0.),
    step_size=EPS,
    max_tree_depth=15,
    max_energy_difference=1000.,
)

chain, _ = nuts_sampler.generate_n_samples(
    SEED, jnp.array(3.), num_samples=N, save_intermediates=True
)
print(f"small mass matrix acceptance: {chain.acceptance}")

ax1.hist(chain.samples, bins=30, density=True)
ax2.plot(chain.samples, linewidth=0.5)

ax1.set_title(rf'$m={1. / nuts_sampler.inverse_mass_matrix:1.2f}$')
ax2.set_title(rf'$m={1. / nuts_sampler.inverse_mass_matrix:1.2f}$')

# %%
nuts_sampler = jft.NUTSChain(
    potential_energy=ham,
    inverse_mass_matrix=50.,
    position_proto=jnp.array(0.),
    step_size=EPS,
    max_tree_depth=15,
    max_energy_difference=1000.,
)

chain, _ = nuts_sampler.generate_n_samples(
    SEED, jnp.array(3.), num_samples=N, save_intermediates=True
)
print(f"large mass matrix acceptance: {chain.acceptance}")

ax3.hist(chain.samples, bins=30, density=True)
ax4.plot(chain.samples, linewidth=0.5)

ax3.set_title(rf'$m={1. / nuts_sampler.inverse_mass_matrix:1.2f}$')
ax4.set_title(rf'$m={1. / nuts_sampler.inverse_mass_matrix:1.2f}$')

# %%
xs = jnp.linspace(-10, 20, num=500)
Z = jnp.trapz(jnp.exp(-ham(xs)), xs)
ax1.plot(xs, jnp.exp(-ham(xs)) / Z, linewidth=0.5, c='r')
ax3.plot(xs, jnp.exp(-ham(xs)) / Z, linewidth=0.5, c='r')

ax1.set_ylabel('frequency')
ax2.set_ylabel('position')
ax3.set_xlabel('position')
ax3.set_ylabel('frequency')
ax4.set_xlabel('time')
ax4.set_ylabel('position')

#fig.suptitle("sum of two Gaussians, with different choices of mass matrix")

fig.tight_layout()
fig.savefig("multimodal.pdf", bbox_inches='tight')
print("final figure saved as multimodal.pdf")



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
from numpy.testing import assert_allclose

from ..common import list2fixture, setup_function, teardown_function

space = list2fixture([
    ift.RGSpace(4),
    ift.HPSpace(4),
    ift.GLSpace(4)
])


def test_const_func(space):
    sig = ift.Field.from_random(domain=space, random_type='normal')
    fco_op = ift.FuncConvolutionOperator(space, lambda x: np.ones(x.shape))
    vals = fco_op(sig).val
    vals = np.round(vals, decimals=5)
    assert len(np.unique(vals)) == 1


def gauss(x, sigma):
    normalization = np.sqrt(2. * np.pi) * sigma
    return np.exp(-0.5 * x * x / sigma**2) / normalization


def test_gaussian_smoothing():
    N = 128
    sigma = N / 10**4
    dom = ift.RGSpace(N)
    sig = ift.Field.from_random(dom, 'normal').ptw("exp")
    fco_op = ift.FuncConvolutionOperator(dom, lambda x: gauss(x, sigma))
    sm_op = ift.HarmonicSmoothingOperator(dom, sigma)
    assert_allclose(fco_op(sig).val,
                    sm_op(sig).val,
                    rtol=1e-05)
    assert_allclose(fco_op.adjoint_times(sig).val,
                    sm_op.adjoint_times(sig).val,
                    rtol=1e-05)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce

from .domain import Domain


class UnstructuredDomain(Domain):
    """A :class:`~nifty8.domains.domain.Domain` subclass for spaces with no
    associated geometry.

    Typically used for data spaces.

    Parameters
    ----------
    shape : tuple of int
        The required shape for an array which can hold the unstructured
        domain's data.
    """

    _needed_for_hash = ["_shape"]

    def __init__(self, shape):
        try:
            self._shape = tuple([int(i) for i in shape])
        except TypeError:
            self._shape = (int(shape), )

    def __repr__(self):
        return "UnstructuredDomain(shape={})".format(self.shape)

    @property
    def shape(self):
        return self._shape

    @property
    def size(self):
        return reduce(lambda x, y: x*y, self.shape)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
from numpy.testing import assert_allclose

from ..common import setup_function, teardown_function


def test_vdot_operator():
    dom = ift.makeDomain(ift.RGSpace(8))
    fa_1 = ift.FieldAdapter(dom, 'f1')
    fa_2 = ift.FieldAdapter(dom, 'f2')
    op1 = fa_1.vdot(fa_2)
    f = ift.from_random(op1.domain, dtype=np.float64)
    arr1 = f['f1'].val
    arr2 = f['f2'].val
    res1 = f['f1'].vdot(f['f2'])
    res2 = op1(f)
    res3 = np.vdot(arr1, arr2)
    assert_allclose(res1.val, res2.val)
    assert_allclose(res1.val, res3)
    ift.extra.check_operator(op1, f)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from types import LambdaType

import nifty8 as ift
import pytest

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize


@pmp('attr_expected_type',
     [['harmonic', bool], ['shape', tuple], ['size', int]])
@pmp('space', [
    ift.RGSpace(4),
    ift.PowerSpace(ift.RGSpace((4, 4), harmonic=True)),
    ift.LMSpace(5),
    ift.HPSpace(4),
    ift.GLSpace(4)
])
def test_property_ret_type(space, attr_expected_type):
    ift.myassert(
        isinstance(
            getattr(space, attr_expected_type[0]), attr_expected_type[1]))


@pmp('method_expected_type',
     [['get_k_length_array', ift.Field],
      ['get_fft_smoothing_kernel_function', 2.0, LambdaType]])
@pmp('space', [ift.RGSpace(4, harmonic=True), ift.LMSpace(5)])
def test_method_ret_type(space, method_expected_type):
    ift.myassert(
        type(
            getattr(space, method_expected_type[0])
            (*method_expected_type[1:-1])) is method_expected_type[-1])



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest

from ..common import setup_function, teardown_function


@pytest.mark.parametrize('sp', [
    ift.RGSpace(4),
    ift.PowerSpace(ift.RGSpace((4, 4), harmonic=True)),
    ift.LMSpace(5),
    ift.HPSpace(4),
    ift.GLSpace(4)
])
@pytest.mark.parametrize('seed', [13, 2])
def test_value_inserter(sp, seed):
    with ift.random.Context(seed):
        ind = tuple([int(ift.random.current_rng().integers(0, ss - 1)) for ss in sp.shape])
        op = ift.ValueInserter(sp, ind)
        f = ift.from_random(op.domain, 'normal')
    inp = f.val
    ret = op(f).val
    ift.myassert(ret[ind] == inp)
    ift.myassert(np.sum(ret) == inp)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import pytest

pytest.importorskip("jax")

import jax.random as random
import pytest
from numpy.testing import assert_allclose

import nifty8 as ift
import nifty8.re as jft

pmp = pytest.mark.parametrize


@pmp("shape", [(2,), (3, 3)])
@pmp("distances", [0.1])
@pmp("offset_mean", [0])
@pmp("offset_std", [(0.1, 0.1)])
@pmp("asperity", [None, (1.0, 1.0)])
@pmp("flexibility", [None, (1.0, 1.0)])
@pmp("fluctuations", [(1.0, 1.0)])
@pmp("loglogavgslope", [(1.0, 1.0)])
def test_correlated_field_non_parametric_init(
    shape,
    distances,
    offset_mean,
    offset_std,
    fluctuations,
    loglogavgslope,
    asperity,
    flexibility,
):
    cf = jft.CorrelatedFieldMaker("cf")
    cf.set_amplitude_total_offset(offset_mean=offset_mean, offset_std=offset_std)
    cf.add_fluctuations(
        shape,
        distances=distances,
        fluctuations=fluctuations,
        loglogavgslope=loglogavgslope,
        asperity=asperity,
        flexibility=flexibility,
    )
    correlated_field = cf.finalize()
    assert correlated_field
    assert correlated_field.domain


@pmp("shape", [(2,), (3, 3)])
@pmp("distances", [0.1])
@pmp("offset_mean", [0])
@pmp("offset_std", [(0.1, 0.1)])
@pmp("scale", [(1.0, 1.0)])
@pmp("loglogslope", [(1.0, 1.0)])
@pmp("cutoff", [(1.0, 1.0)])
def test_correlated_field_matern_init(
    shape,
    distances,
    offset_mean,
    offset_std,
    scale,
    loglogslope,
    cutoff,
):
    cf = jft.CorrelatedFieldMaker("cf")
    cf.set_amplitude_total_offset(offset_mean=offset_mean, offset_std=offset_std)
    cf.add_fluctuations_matern(
        shape,
        distances=distances,
        scale=scale,
        cutoff=cutoff,
        loglogslope=loglogslope,
        renormalize_amplitude=False,
    )
    correlated_field = cf.finalize()
    assert correlated_field
    assert correlated_field.domain


@pmp("flu", ([1e-1], [1e-1, 5e-3], [1e-1, 5e-3, 5e-3], 1e-1))
@pmp("slp", ([1e-1], [1e-1, 5e-3], [1e-1, 5e-3, 5e-3], 1e-1))
@pmp("flx", ([1e-1], [1e-1, 5e-3], [1e-1, 5e-3, 5e-3], 1e-1))
@pmp("asp", ([1e-1], [1e-1, 5e-3], [1e-1, 5e-3, 5e-3], 1e-1))
def test_correlated_field_non_parametric_init_validation(flu, slp, flx, asp):
    dims = (16,)
    if all(
        (
            isinstance(el, (tuple, list))
            and len(el) == 2
            and all(isinstance(f, float) for f in el)
        )
        for el in (flu, slp, flx, asp)
    ):
        return
    with pytest.raises(TypeError):
        cf_zm = dict(offset_mean=0.0, offset_std=(1e-3, 1e-4))
        cf_fl = dict(
            fluctuations=flu,
            loglogavgslope=slp,
            flexibility=flx,
            asperity=asp,
        )
        cfm = jft.CorrelatedFieldMaker("cf")
        cfm.set_amplitude_total_offset(**cf_zm)
        cfm.add_fluctuations(
            shape=dims,
            distances=tuple(1.0 / d for d in dims),
            **cf_fl,
            prefix="ax1",
            non_parametric_kind="power",
        )
        cfm.finalize()


@pmp("seed", [0, 42])
@pmp("ht_convention", ["canonical_hartley", "non_canonical_hartley"])
@pmp(
    "shape,distances,harmonic_type",
    [
        ((4,), 0.1, "fourier"),
        ((4,), 5.0, "fourier"),
        ((3, 3), 0.1, "fourier"),
        ((3, 3), 5.0, "fourier"),
        ((4,), None, "spherical"),
        ((32,), None, "spherical"),
    ],
)
@pmp("offset_mean", [0])
@pmp("offset_std", [(0.1, 0.1)])
@pmp("fluctuations", [(1.0, 0.1), (3.0, 2.0)])
@pmp("loglogavgslope", [(-1.0, 0.1), (4.0, 1.0)])
@pmp("flexibility", [(1.0, 0.1), (3.0, 2.0)])
@pmp("asperity", [None, (0.2, 2.0e-2)])
def test_nifty_vs_niftyre_non_parametric_cf(
    seed,
    ht_convention,
    shape,
    distances,
    harmonic_type,
    offset_mean,
    offset_std,
    fluctuations,
    loglogavgslope,
    asperity,
    flexibility,
):
    if harmonic_type == "spherical":
        try:
            from jaxbind.contrib import jaxducc0
        except (ImportError, ModuleNotFoundError):
            pytest.skip("skipping since `jaxducc0` from `jaxbind` is not available")
    jft.config.update("hartley_convention", ht_convention)

    key = random.PRNGKey(seed)
    fluct_kwargs = dict(
        fluctuations=fluctuations,
        loglogavgslope=loglogavgslope,
        asperity=asperity,
        flexibility=flexibility,
    )

    jcfm = jft.CorrelatedFieldMaker("")
    jcfm.set_amplitude_total_offset(offset_mean=offset_mean, offset_std=offset_std)
    jcfm.add_fluctuations(
        shape,
        distances=distances,
        **fluct_kwargs,
        non_parametric_kind="power",
        harmonic_type=harmonic_type,
    )
    jcf = jcfm.finalize()

    cfm = ift.CorrelatedFieldMaker("")
    cfm.set_amplitude_total_offset(offset_mean, offset_std)
    sp = (
        ift.RGSpace(shape, distances)
        if distances is not None
        else ift.HPSpace(shape[0])
    )
    cfm.add_fluctuations(sp, **fluct_kwargs)
    cf = cfm.finalize(prior_info=0)

    pos = jft.random_like(key, jcf.domain)
    npos = {
        k: ift.makeField(cf.domain[k], v if k != "spectrum" else v.T)
        for k, v in pos.items()
    }
    npos = ift.MultiField.from_dict(npos, cf.domain)
    assert_allclose(cf(npos).val, jcf(pos))


@pmp("seed", [0, 42])
@pmp("shape", [(2,), (3, 3)])
@pmp("distances", [0.1, 5.0])
@pmp("offset_mean", [0])
@pmp("offset_std", [(0.1, 0.1)])
@pmp("scale", [(1.0, 1.0), (3.0, 2.0)])
@pmp("loglogslope", [(1.0, 1.0), (5.0, 0.5)])
@pmp("cutoff", [(1.0, 1.0), (0.1, 0.01)])
def test_nifty_vs_niftyre_matern_cf(
    seed, shape, distances, offset_mean, offset_std, scale, cutoff, loglogslope
):
    key = random.PRNGKey(seed)
    fluct_kwargs = dict(scale=scale, cutoff=cutoff, loglogslope=loglogslope)

    jcfm = jft.CorrelatedFieldMaker("")
    jcfm.set_amplitude_total_offset(offset_mean=offset_mean, offset_std=offset_std)
    jcfm.add_fluctuations_matern(
        shape,
        distances=distances,
        **fluct_kwargs,
        non_parametric_kind="amplitude",
        renormalize_amplitude=False,
    )
    jcf = jcfm.finalize()

    cfm = ift.CorrelatedFieldMaker("")
    cfm.set_amplitude_total_offset(offset_mean, offset_std)
    cfm.add_fluctuations_matern(ift.RGSpace(shape, distances), **fluct_kwargs)
    cf = cfm.finalize(prior_info=0)

    pos = jft.random_like(key, jcf.domain)
    npos = {k: ift.makeField(cf.domain[k], v) for k, v in pos.items()}
    npos = ift.MultiField.from_dict(npos, cf.domain)
    assert_allclose(cf(npos).val, jcf(pos))


CFG_OFFSET = dict(offset_mean=0, offset_std=(0.1, 0.1))
CFG_FLUCT = dict(
    fluctuations=(1.0, 0.1),
    loglogavgslope=(-1.0, 0.1),
    flexibility=(1.0, 0.1),
    asperity=(0.2, 2.0e-2),
)
FLUCTUATIONS_CHOICES = (
    dict(shape=(4,), distances=None, harmonic_type="spherical"),
    dict(shape=(3, 3), distances=(0.1, 0.1), harmonic_type="fourier"),
    dict(shape=(6,), distances=(1.0,), harmonic_type="fourier"),
)


@pmp("seed", [0, 42])
@pmp("fluct1", FLUCTUATIONS_CHOICES)
@pmp("fluct2", FLUCTUATIONS_CHOICES)
def test_nifty_vs_niftyre_product(seed, fluct1, fluct2):
    if any(f["harmonic_type"] == "spherical" for f in (fluct1, fluct2)):
        try:
            from jaxbind.contrib import jaxducc0
        except (ImportError, ModuleNotFoundError):
            pytest.skip("skipping since `jaxducc0` from `jaxbind` is not available")
    key = random.PRNGKey(seed)

    jcfm = jft.CorrelatedFieldMaker("")
    jcfm.set_amplitude_total_offset(**CFG_OFFSET)
    for i, f in enumerate((fluct1, fluct2)):
        jcfm.add_fluctuations(
            **f, **CFG_FLUCT, non_parametric_kind="power", prefix=f"space{i}"
        )
    jcf = jcfm.finalize()

    cfm = ift.CorrelatedFieldMaker("")
    cfm.set_amplitude_total_offset(**CFG_OFFSET)
    for i, f in enumerate((fluct1, fluct2)):
        sp = (
            ift.HPSpace(f["shape"][0])
            if f["distances"] is None
            else ift.RGSpace(f["shape"], f["distances"])
        )
        cfm.add_fluctuations(sp, **CFG_FLUCT, prefix=f"space{i}")
    cf = cfm.finalize(prior_info=0)

    pos = jft.random_like(key, jcf.domain)
    npos = {
        k: ift.makeField(cf.domain[k], v if not k.endswith("spectrum") else v.T)
        for k, v in pos.items()
    }
    npos = ift.MultiField.from_dict(npos, cf.domain)
    assert_allclose(cf(npos).val, jcf(pos))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from ..common import setup_function, teardown_function

pmp = pytest.mark.parametrize
# [shape, distances, harmonic, expected]
CONSTRUCTOR_CONFIGS = [[(8,), None, False, {
    'shape': (8,),
    'distances': (0.125,),
    'harmonic': False,
    'size': 8,
}], [(8,), None, True, {
    'shape': (8,),
    'distances': (1.0,),
    'harmonic': True,
    'size': 8,
}], [(8,), (12,), True, {
    'shape': (8,),
    'distances': (12.0,),
    'harmonic': True,
    'size': 8,
}], [(11, 11), None, False, {
    'shape': (11, 11),
    'distances': (1/11, 1/11),
    'harmonic': False,
    'size': 121,
}], [(12, 12), (1.3, 1.3), True, {
    'shape': (12, 12),
    'distances': (1.3, 1.3),
    'harmonic': True,
    'size': 144,
}]]


def get_k_length_array_configs():
    # for RGSpace(shape=(4, 4), distances=(0.25,0.25))
    cords_0 = np.ogrid[0:4, 0:4]
    da_0 = ((cords_0[0] - 4//2)*0.25)**2
    da_0 = np.fft.ifftshift(da_0)
    temp = ((cords_0[1] - 4//2)*0.25)**2
    temp = np.fft.ifftshift(temp)
    da_0 = da_0 + temp
    da_0 = np.sqrt(da_0)
    return [
        [(4, 4), (0.25, 0.25), da_0],
    ]


def get_dvol_configs():
    return [[(11, 11), None, False, 1], [(11, 11), None, False, 1],
            [(11, 11), (1.3, 1.3), False, 1], [(11, 11), (1.3, 1.3), False,
                                               1], [(11, 11), None, True, 1],
            [(11, 11), None, True, 1], [(11, 11), (1.3, 1.3), True,
                                        1], [(11, 11), (1.3, 1.3), True, 1]]


@pmp('attribute, expected_type', [['distances', tuple]])
def test_property_ret_type(attribute, expected_type):
    x = ift.RGSpace(1)
    ift.myassert(isinstance(getattr(x, attribute), expected_type))


@pmp('shape, distances, harmonic, expected', CONSTRUCTOR_CONFIGS)
def test_constructor(shape, distances, harmonic, expected):
    x = ift.RGSpace(shape, distances, harmonic)
    for key, value in expected.items():
        assert_equal(getattr(x, key), value)


@pmp('shape, distances, expected', get_k_length_array_configs())
def test_k_length_array(shape, distances, expected):
    r = ift.RGSpace(shape=shape, distances=distances, harmonic=True)
    assert_allclose(r.get_k_length_array().val, expected)


@pmp('shape, distances, harmonic, power', get_dvol_configs())
def test_dvol(shape, distances, harmonic, power):
    r = ift.RGSpace(shape=shape, distances=distances, harmonic=harmonic)
    assert_allclose(r.dvol, np.prod(r.distances)**power)


def test_codomain():
    for i in range(1, 1000):
        r = ift.RGSpace(shape=(i,), distances=(1.,), harmonic=False)
        assert_equal(r.get_default_codomain().get_default_codomain(), r)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.


import operator

import numpy as np
import scipy.fft

from .config import _config

_nthreads = 1


def nthreads():
    return _nthreads


def set_nthreads(nthr):
    global _nthreads
    _nthreads = int(nthr)


def _scipy_fftn(a, axes=None):
    return scipy.fft.fftn(a, axes=axes, workers=_nthreads)


def _scipy_ifftn(a, axes=None):
    return scipy.fft.ifftn(a, axes=axes, workers=_nthreads)


def _scipy_hartley(a, axes=None):
    tmp = scipy.fft.fftn(a, axes=axes, workers=_nthreads)
    c = _config.get("hartley_convention")
    add_or_sub = operator.add if c == "non_canonical_hartley" else operator.sub
    return add_or_sub(tmp.real, tmp.imag)


def _scipy_vdot(a, b):
    from .logger import logger
    if (isinstance(a, np.ndarray) and a.dtype == np.float32) or \
    (isinstance(b, np.ndarray) and b.dtype == np.float32):
        logger.warning("Calling np.vdot in single precision may lead to inaccurate results")
    return np.vdot(a, b)


try:
    import ducc0.fft as my_fft
    import ducc0.misc


    def fftn(a, axes=None):
        return my_fft.c2c(a, axes=axes, nthreads=max(_nthreads, 0))


    def ifftn(a, axes=None):
        return my_fft.c2c(a, axes=axes, inorm=2, forward=False,
                          nthreads=max(_nthreads, 0))


    def hartley(a, axes=None):
        c = _config.get("hartley_convention")
        ht = my_fft.genuine_hartley if c == "non_canonical_hartley" else my_fft.genuine_fht
        return ht(a, axes=axes, nthreads=max(_nthreads, 0))


    def vdot(a, b):
        if isinstance(a, np.ndarray) and np.issubdtype(a.dtype, np.integer):
            a = a.astype(np.float64)
        if isinstance(b, np.ndarray) and np.issubdtype(b.dtype, np.integer):
            b = b.astype(np.float64)
        return ducc0.misc.vdot(a, b)

except ImportError:
    fftn = _scipy_fftn
    ifftn = _scipy_ifftn
    hartley = _scipy_hartley
    vdot = _scipy_vdot



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from typing import Any, Callable, Dict, Hashable, Mapping, TypeVar

import jax
from jax import numpy as jnp

O = TypeVar('O')
I = TypeVar('I')


def hvp(f, primals, tangents):
    return jax.jvp(jax.grad(f), primals, tangents)[1]


def split(mappable, keys):
    """Split a dictionary into one containing only the specified keys and one
    with all of the remaining ones.
    """
    sel, rest = {}, {}
    for k, v in mappable.items():
        if k in keys:
            sel[k] = v
        else:
            rest[k] = v
    return sel, rest


def isiterable(candidate):
    try:
        iter(candidate)
        return True
    except (TypeError, AttributeError):
        return False


def is_iterable_of_non_iterables(ls: Any) -> bool:
    """Indicates whether the input is one dimensional.

    An object is considered one dimensional if it is an iterable of
    non-iterable items.
    """
    if hasattr(ls, "ndim"):
        return ls.ndim == 1
    if not isiterable(ls):
        return False
    return all(not isiterable(e) for e in ls)


def doc_from(original):
    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper


def wrap(
    call: Callable[[I], O],
    name: Hashable,
) -> Callable[[Mapping[Hashable, I]], O]:
    def named_call(p):
        return call(p[name])

    return named_call


def wrap_left(
    call: Callable[[I], O],
    name: Hashable,
) -> Callable[[I], Dict[Hashable, O]]:
    def named_call(p):
        return {name: call(p)}

    return named_call


def interpolate(xmin=-7., xmax=7., N=14000) -> Callable:
    """Replaces a local nonlinearity such as jnp.exp with a linear interpolation

    Interpolating functions speeds up code and increases numerical stability in
    some cases, but at a cost of precision and range.

    Parameters
    ----------
    xmin : float
        Minimal interpolation value. Default: -7.
    xmax : float
        Maximal interpolation value. Default: 7.
    N : int
        Number of points used for the interpolation. Default: 14000
    """
    def decorator(f):
        from functools import wraps

        x = jnp.linspace(xmin, xmax, N)
        y = f(x)

        @wraps(f)
        def wrapper(t):
            return jnp.interp(t, x, y)

        return wrapper

    return decorator



import sys

import pytest
pytest.importorskip("jax")

from jax import numpy as jnp
from jax.scipy import stats
from numpy.testing import assert_allclose
import scipy
from scipy.special import comb

import nifty8.re as jft

pmp = pytest.mark.parametrize


def mnc2mc(mnc, wmean=True):
    """Convert non-central to central moments, uses recursive formula
    optionally adjusts first moment to return mean.
    """

    # https://www.statsmodels.org/stable/_modules/statsmodels/stats/moment_helpers.html
    def _local_counts(mnc):
        mean = mnc[0]
        mnc = [1] + list(mnc)  # add zero moment = 1
        mu = []
        for n, m in enumerate(mnc):
            mu.append(0)
            for k in range(n + 1):
                sgn_comb = (-1)**(n - k) * comb(n, k, exact=True)
                mu[n] += sgn_comb * mnc[k] * mean**(n - k)
        if wmean:
            mu[1] = mean
        return mu[1:]

    res = jnp.apply_along_axis(_local_counts, 0, mnc)
    # for backward compatibility convert 1-dim output to list/tuple
    return res


# Test simple distributions with no extra parameters
dists = [stats.cauchy, stats.expon, stats.laplace, stats.logistic, stats.norm]
# Tuple of `rtol` and `atol` for every tested moment
moments_tol = {1: (0., 2e-1), 2: (3e-1, 0.), 3: (4e-1, 8e-1), 4: (4., 0.)}


@pmp("distribution", dists)
def test_moment_consistency(distribution, plot=False):
    name = distribution.__name__.split('.')[-1]

    max_tree_depth = 20
    sampler = jft.NUTSChain(
        potential_energy=lambda x: -1 * distribution.logpdf(x),
        inverse_mass_matrix=1.,
        position_proto=jnp.array(0.),
        step_size=0.7193,
        max_tree_depth=max_tree_depth,
    )
    chain, _ = sampler.generate_n_samples(
        42, jnp.array(1.03890), num_samples=1000, save_intermediates=True
    )

    # unique, counts = jnp.unique(chain.depths, return_counts=True)
    # depths_frequencies = jnp.asarray((unique, counts)).T

    if plot is True:
        import matplotlib.pyplot as plt

        fig, axs = plt.subplots(1, 2)

        bins = jnp.linspace(-10, 10)
        if distribution is stats.expon:
            bins = jnp.linspace(0, 10)
        axs.flat[0].hist(
            chain.samples, bins=bins, density=True, histtype="step"
        )
        axs.flat[0].plot(bins, distribution.pdf(bins), color='r')
        axs.flat[0].set_title(f"{name} PDF")

        axs.flat[1].hist(
            chain.depths,
            bins=jnp.arange(max_tree_depth + 1),
            density=True,
            histtype="step"
        )
        axs.flat[1].set_title(f"Tree-depth")
        fig.tight_layout()
        plt.show()

    # central moments; except for the first (i.e. mean)
    sample_moms_central = scipy.stats.moment(chain.samples, [1, 2, 3, 4, 5, 6])
    sample_moms_central[0] = jnp.mean(chain.samples)

    scipy_dist = getattr(scipy.stats, name)
    dist_moms_non_central = jnp.array(
        [scipy_dist.moment(i) for i in [1, 2, 3, 4, 5, 6]]
    )
    dist_moms_central = mnc2mc(dist_moms_non_central, wmean=True)

    for i, (smpl_mom, dist_mom) in enumerate(
        zip(sample_moms_central, dist_moms_central), start=1
    ):
        msg = (
            f"{name} (moment {i}) :: sampled: {smpl_mom:+.2e}"
            f" true: {dist_mom:+.2e} tested: "
        )
        print(msg, end="", file=sys.stderr)
        test = not jnp.isnan(dist_mom)
        test &= not (jnp.allclose(dist_mom, 0.) and i > 1)
        if i in moments_tol and test:
            assert_allclose(
                dist_mom, smpl_mom,
                **dict(zip(("rtol", "atol"), moments_tol[i]))
            )
            print("✓", file=sys.stderr)
        else:
            print("✗", file=sys.stderr)


if __name__ == "__main__":
    for d in dists:
        test_moment_consistency(d, plot=True)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
# Author: Philipp Arras

import nifty8 as ift
import pytest
from mpi4py import MPI

from ..common import list2fixture, setup_function, teardown_function

pmp = pytest.mark.parametrize
comm = [MPI.COMM_WORLD]
if MPI.COMM_WORLD.Get_size() == 1:
    comm += [None]
comm = list2fixture(comm)


def _get_sample_list(communicator, cls):
    dom = ift.makeDomain({"a": ift.UnstructuredDomain(2), "b": ift.RGSpace(12)})
    samples = [ift.from_random(dom) for _ in range(3)]
    mean = ift.from_random(dom)
    if cls == "SampleList":
        return ift.SampleList(samples, communicator), samples
    elif cls == "ResidualSampleList":
        neg = 3*[False]
        return ift.ResidualSampleList(mean, samples, neg, communicator), [mean + ss for ss in samples]
    elif cls == "SymmetricalSampleList":
        neg = len(samples)*[False] + len(samples)*[True]
        reference = [mean + ss for ss in samples] + [mean - ss for ss in samples]
        samples = samples + samples
        return ift.ResidualSampleList(mean, samples, neg, communicator), reference
    elif cls == "PartiallyEmptyResidualSampleList":
        if communicator is None or communicator.Get_rank() == 0:
            neg = [False]
            resi = samples[0:1]
        else:
            neg = resi = []
        return ift.ResidualSampleList(mean, resi, neg, communicator), [mean + samples[0]]
    elif cls == "PartiallyEmptySampleList":
        if communicator is None or communicator.Get_rank() == 0:
            local_samples = samples[0:1]
        else:
            local_samples = []
        return ift.SampleList(local_samples, comm=communicator, domain=samples[0].domain), samples[0:1]
    raise NotImplementedError


all_cls = ["ResidualSampleList", "SampleList", "PartiallyEmptyResidualSampleList",
           "PartiallyEmptySampleList"]


def _get_ops(sample_list):
    dom = sample_list.domain
    return [None,
            ift.ScalingOperator(dom, 1.),
            ift.ducktape(None, dom, "a") @ ift.ScalingOperator(dom, 1.).exp()]


@pmp("cls", ["SampleList", "PartiallyEmptySampleList"])
def test_sample_list(comm, cls):
    sl, samples = _get_sample_list(comm, cls)
    assert comm == sl.comm

    for op in _get_ops(sl):
        sc = ift.StatCalculator()
        if op is None:
            [sc.add(ss) for ss in samples]
        else:
            [sc.add(op(ss)) for ss in samples]
        if sl.n_samples > 1:
            mean, var = sl.sample_stat(op)
            ift.extra.assert_allclose(mean, sl.average(op))
            ift.extra.assert_allclose(mean, sc.mean)  # FIXME Why does this not fail for comm != None?
            if comm is None:
                ift.extra.assert_allclose(var, sc.var)

        samples = list(samples)
        if op is not None:
            samples = [op(ss) for ss in samples]

        for s0, s1 in zip(samples, sl.iterator(op)):
            ift.extra.assert_equal(s0, s1)

        if comm is None:
            assert len(samples) == sl.n_samples
        else:
            assert len(samples) <= sl.n_samples


@pmp("cls", all_cls)
def test_load_and_save(comm, cls):
    if comm is None and ift.utilities.get_MPI_params()[1] > 1:
        pytest.skip()

    sl, _ = _get_sample_list(comm, cls)
    sl.save("sl")
    lcls = cls[14:] if cls[:14] == "PartiallyEmpty" else cls
    sl1 = getattr(ift, lcls).load("sl", comm)

    for s0, s1 in zip(sl.local_iterator(), sl1.local_iterator()):
        ift.extra.assert_equal(s0, s1)

    for s0, s1 in zip(sl.local_iterator(), sl1.local_iterator()):
        ift.extra.assert_equal(s0, s1)


def test_load_mean(comm):
    sl, _ = _get_sample_list(comm, "SymmetricalSampleList")
    m0 = sl._m
    sl.save("sl")
    sl1 = ift.ResidualSampleList.load("sl", comm=comm)
    m1, _ = sl1.sample_stat(None)
    m2 = ift.ResidualSampleList.load_mean("sl")
    ift.extra.assert_equal(m0, m2)
    ift.extra.assert_allclose(m1, m2)


@pmp("cls", all_cls)
@pmp("mean", [False, True])
@pmp("std", [False, True])
@pmp("samples", [False, True])
def test_save_to_hdf5(comm, cls, mean, std, samples):
    pytest.importorskip("h5py")
    import h5py
    import numpy as np
    from nifty8 import DomainTuple, MultiDomain, RGSpace, UnstructuredDomain

    if comm is None and ift.utilities.get_MPI_params()[1] > 1:
        pytest.skip()
    sl, _ = _get_sample_list(comm, cls)
    if sl.n_samples < 2 and std:
        pytest.skip()
    for op in _get_ops(sl):
        if not mean and not std and not samples:
            with pytest.raises(ValueError):
                sl.save_to_hdf5("output.h5", op, mean=mean, std=std, samples=samples)
            continue
        sl.save_to_hdf5("output.h5", op, mean=mean, std=std, samples=samples, overwrite=True)
        if comm is not None:
            comm.Barrier()

        flddom = sl.domain if op is None else op.target
        mdom = isinstance(flddom, ift.MultiDomain)

        shps = []
        f = h5py.File("output.h5", "r")
        domain_repr = f.attrs["nifty domain"]
        if mean:
            shps.append(_get_shape(f["stats"]["mean"], mdom))
        if std:
            shps.append(_get_shape(f["stats"]["standard deviation"], mdom))
        if samples:
            for ii in range(sl.n_samples):
                shps.append(_get_shape(f["samples"][str(ii)], mdom))
        assert all(elem == shps[0] for elem in shps)

        dom1 = eval(domain_repr)
        assert dom1 is flddom

        f.close()


def _get_shape(inp, mdom):
    if not mdom:
        return inp.shape
    return {kk: _get_shape(vv, False) for kk, vv in inp.items()}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import pytest
from mpi4py import MPI

from ..common import setup_function, teardown_function

comm = MPI.COMM_WORLD
ntask = comm.Get_size()
rank = comm.Get_rank()
master = (rank == 0)
mpi = ntask > 1

pmp = pytest.mark.parametrize
pms = pytest.mark.skipif


@pms(not mpi, reason="requires at least two mpi tasks")
def test_MPI_equality():
    obj = rank
    with pytest.raises(RuntimeError):
        ift.utilities.check_MPI_equality(obj, comm)

    obj = [ii + rank for ii in range(10, 12)]
    with pytest.raises(RuntimeError):
        ift.utilities.check_MPI_equality(obj, comm)

    sseqs = ift.random.spawn_sseq(ntask)
    for obj in [12., None, (29, 30), [1, 2, 3], sseqs[0], sseqs]:
        ift.utilities.check_MPI_equality(obj, comm)

    obj = ift.random.spawn_sseq(ntask, parent=sseqs[comm.rank])
    with pytest.raises(RuntimeError):
        ift.utilities.check_MPI_equality(obj, comm)


@pms(not mpi, reason="requires at least two mpi tasks")
def test_MPI_synced_random_state():
    ift.utilities.check_MPI_synced_random_state(comm)
    with ift.random.Context(123 if master else 111):
        with pytest.raises(RuntimeError):
            ift.utilities.check_MPI_synced_random_state(comm)


@pms(not mpi, reason="requires at least two mpi tasks")
@pmp("geo", [False, True])
@pmp("mirror", [False, True])
@pmp("n_samples", [2, 3])
def test_MPI_synced_random_state_kl_energies(geo, mirror, n_samples):
    ic = ift.AbsDeltaEnergyController(0.1, iteration_limit=2)
    lh = ift.GaussianEnergy(ift.full(ift.UnstructuredDomain(2), 2.)).ducktape("a")
    ham = ift.StandardHamiltonian(lh, ic)
    ift.utilities.check_MPI_synced_random_state(comm)
    with ift.random.Context(123 if master else 111):
        mean = ift.from_random(ham.domain)
        with pytest.raises(RuntimeError):
            mini = None
            if geo:
                mini = ift.NewtonCG(ift.AbsDeltaEnergyController(0.1, iteration_limit=2))
            ift.SampledKLEnergy(mean, ham, n_samples, mini, comm=comm)


@pms(not mpi, reason="requires at least two mpi tasks")
@pmp("sync", [False, True])
def test_random_field_generation(sync):
    with ift.random.Context(123 if master and not sync else 111):
        dom = ift.UnstructuredDomain(5)
        fld = ift.from_random(dom)
        if sync:
            ift.utilities.check_MPI_equality(fld, comm)
        else:
            with pytest.raises(RuntimeError):
                ift.utilities.check_MPI_equality(fld, comm)



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from typing import Callable, Optional, Tuple, Union

import jax
from jax import numpy as jnp
import numpy as np

from .correlated_field import get_fourier_mode_distributor, hartley

NDArray = Union[jnp.ndarray, np.ndarray]


def matmul_toeplitz(c, x):
    """Efficiently compute the product of an implicit Toeplitz matrix with a
    vector.

    Parameters
    ----------
    c : ndarray
        One-dimensional column of the Toeplitz matrix. It is assumed that the
        row `r` is given by `r = conjugate(c)`.
    x : (M,) or (M, K) ndarray
        Vector or matrix which to multiply from the right to the Toeplitz
        matrix.
    Returns
    -------
    Toeplitz @ x : (M,) or (M, K) ndarray
        Result of the matrix multiplication.
    """
    from jax.numpy.fft import fft, ifft, rfft, irfft

    c = c.ravel()
    r = c.conjugate()

    toepl_shp = (len(c), len(c))
    x_shp = x.shape
    if x.shape[0] != r.shape[0] or x.ndim > 2:
        raise ValueError('invalid matrix product dimensions')
    x = x.reshape(-1, 1) if x.ndim == 1 else x.reshape(x.shape[0], -1)
    m = x.shape[1]

    dtp = np.result_type(*(i.dtype for i in (r, c, x)))
    r, c, x = (jnp.asarray(i, dtype=dtp) for i in (r, c, x))
    embedded_col = jnp.concatenate((c, r[-1:0:-1]))

    # For real inputs use rfft
    cmplx = jnp.iscomplexobj(embedded_col) or jnp.iscomplexobj(x)
    ht, iht = (fft, ifft) if cmplx else (rfft, irfft)
    p = toepl_shp[0] + toepl_shp[1] - 1
    ht_mat = ht(embedded_col, axis=0).reshape(-1, 1)
    ht_x = ht(x, n=p, axis=0)

    toepl_times_x = iht(ht_mat * ht_x, axis=0, n=p)[:toepl_shp[0], :]

    out_shp = (toepl_shp[0], ) if len(x_shp) == 1 else (toepl_shp[0], m)
    return toepl_times_x.reshape(*out_shp)


def interp_mat(grid_shape, grid_bounds, sampling_points, *, distances=None):
    from scipy.sparse import coo_matrix  # TODO: use only JAX w/o SciPy or NumPy
    from jax.experimental.sparse import BCOO

    if sampling_points.ndim != 2:
        ve = f"invalid dimension of sampling_points {sampling_points.ndim!r}"
        raise ValueError(ve)
    ndim, n_points = sampling_points.shape
    if grid_bounds is not None and len(grid_bounds) != ndim:
        ve = (
            f"grid_bounds of length {len(grid_bounds)} incompatible with"
            f" sampling_points of shape {sampling_points.shape!r}"
        )
        raise ValueError(ve)
    elif grid_bounds is not None:
        offset = np.array(list(zip(*grid_bounds))[0])
    else:
        offset = np.zeros(ndim)
    if distances is not None and np.size(distances) != ndim:
        ve = (
            f"distances of size {np.size(distances)} incompatible with"
            f" sampling_points of shape {sampling_points.shape!r}"
        )
        raise ValueError(ve)
    distances = np.asarray(distances) if distances is not None else None
    if (distances is not None and grid_bounds
        is not None) or (distances is None and grid_bounds is None):
        raise ValueError("exactly one of `distances` or `grid_shape` expected")
    elif grid_bounds is not None:
        distances = np.array(
            [(b[1] - b[0]) / sz for b, sz in zip(grid_bounds, grid_shape)]
        )
    if distances is None:
        raise AssertionError()

    mg = np.mgrid[(slice(0, 2), ) * ndim].reshape(ndim, -1)
    pos = (sampling_points - offset.reshape(-1, 1)) / distances.reshape(-1, 1)
    excess, pos = np.modf(pos)
    pos = pos.astype(np.int64)
    # max_index = np.array(grid_shape).reshape(-1, 1)
    weights = np.zeros((2**ndim, n_points))
    ii = np.zeros((2**ndim, n_points), dtype=np.int64)
    jj = np.zeros((2**ndim, n_points), dtype=np.int64)
    for i in range(2**ndim):
        weights[i, :] = np.prod(
            np.abs(1 - mg[:, i].reshape(-1, 1) - excess), axis=0
        )
        fromi = (pos + mg[:, i].reshape(-1, 1))  # % max_index
        ii[i, :] = np.arange(n_points)
        jj[i, :] = np.ravel_multi_index(fromi, grid_shape)

    mat = coo_matrix(
        (weights.ravel(), (ii.ravel(), jj.ravel())),
        shape=(n_points, np.prod(grid_shape))
    )
    # BCOO(
    #     (weights.ravel(), jnp.stack((ii.ravel(), jj.ravel()), axis=1)),
    #     shape=(n_points, np.prod(grid_shape))
    # )
    return BCOO.from_scipy_sparse(mat).sort_indices()


class HarmonicSKI():
    def __init__(
        self,
        grid_shape: Tuple[int],
        grid_bounds: Tuple[Tuple[float, float]],
        sampling_points: NDArray,
        harmonic_kernel: Optional[Callable] = None,
        padding: float = 0.5,
        subslice=None,
        jitter: Union[bool, float, None] = True
    ):
        """Instantiate a KISS-GP model of the covariance using a harmonic
        representation of the kernel.

        Parameters
        ----------
        grid_shape :
            Number of pixels along each axes of the inducing points within
            `grid_bounds`.
        grid_bounds :
            Tuple of boundaries of length of the number of dimensions. The
            boundaries should denote the leftmost and rightmost edge of the
            modeling space.
        sampling_points :
            Locations of the modeled points within the grid.
        harmonic_kernel :
            Harmonically transformed kernel.
        padding :
            Padding factor which to apply along each axis.
        subslice :
            Slice of the inducing points which to use to model
            `sampling_points`. By default, the subslice is determined by the
            padding.
        jitter :
            Strength of the diagonal jitter which to add to the covariance.
        """
        if jitter is True:
            if sampling_points.dtype.type == np.float64:
                self.jitter = 1e-8
            elif sampling_points.dtype.type == np.float32:
                self.jitter = 1e-6
            else:
                raise NotImplementedError()
        elif jitter is False:
            self.jitter = None
        else:
            self.jitter = jitter

        self.grid_unpadded_shape = np.asarray(grid_shape)
        self.grid_unpadded_bounds = np.asarray(grid_bounds)
        self.grid_unpadded_distances = jnp.diff(
            self.grid_unpadded_bounds, axis=1
        ).ravel() / self.grid_unpadded_shape
        self.grid_unpadded_total_volume = jnp.prod(
            self.grid_unpadded_shape * self.grid_unpadded_distances
        )
        self.w = interp_mat(grid_shape, grid_bounds, sampling_points)

        if padding is not None and padding != 0.:
            pad = 1. + padding
            grid_shape = np.asarray(grid_shape)
            grid_shape_wpad = np.ceil(grid_shape * pad).astype(int)
            scl = grid_shape_wpad / grid_shape
            p = jnp.diff(jnp.asarray(grid_bounds), axis=1).ravel() * (1. - scl)
            grid_bounds_wpad = jnp.asarray(grid_bounds)
            grid_bounds_wpad = grid_bounds_wpad.at[:, 0].set(
                grid_bounds_wpad[:, 0].ravel() - p
            )
            grid_bounds_wpad = grid_bounds_wpad.at[:, 1].set(
                grid_bounds_wpad[:, 1].ravel() + p
            )
            if subslice is None:
                subslice = tuple(map(int, grid_shape))
            grid_shape = grid_shape_wpad
            grid_bounds = grid_bounds_wpad
        self.grid_shape = np.asarray(grid_shape)
        self.grid_bounds = np.asarray(grid_bounds)
        self.grid_distances = jnp.diff(self.grid_bounds,
                                       axis=1).ravel() / self.grid_shape
        self.grid_total_volume = jnp.prod(self.grid_shape * self.grid_distances)

        self.power_distributor, self.unique_mode_lengths, _ = get_fourier_mode_distributor(
            self.grid_shape, self.grid_distances
        )

        if subslice is not None:
            if isinstance(subslice, slice):
                subslice = (subslice, ) * len(self.grid_shape)
            elif isinstance(subslice, int):
                subslice = (slice(subslice), ) * len(self.grid_shape)
            elif isinstance(subslice, tuple):
                if all(isinstance(el, slice) for el in subslice):
                    pass
                elif all(isinstance(el, int) for el in subslice):
                    subslice = tuple(slice(el) for el in subslice)
                else:
                    raise TypeError("elements of `subslice` of invalid type")
            else:
                raise TypeError("`subslice` of invalid type")
        self.grid_subslice = subslice

        self._harmonic_kernel = harmonic_kernel

    @property
    def harmonic_kernel(self) -> Callable:
        """Yields the harmonic kernel specified during initialization or throw
        a `TypeError`.
        """
        if self._harmonic_kernel is None:
            te = (
                "either specify a fixed harmonic kernel during initialization"
                f" of the {self.__class__.__name__} class or provide one here"
            )
            raise TypeError(te)
        return self._harmonic_kernel

    def power(self, harmonic_kernel=None) -> NDArray:
        if harmonic_kernel is None:
            harmonic_kernel = self.harmonic_kernel
        power = harmonic_kernel(self.unique_mode_lengths)
        power *= self.grid_total_volume / self.grid_unpadded_total_volume
        return power

    def amplitude(self, harmonic_kernel=None):
        power = self.power(harmonic_kernel)
        # Assume that the kernel scales linear with the total volume
        return jnp.sqrt(power)

    def harmonic_transform(self, x) -> NDArray:
        return 1. / self.grid_total_volume * hartley(x)

    def correlated_field(self, x, harmonic_kernel=None) -> NDArray:
        amp = self.amplitude(harmonic_kernel)
        f = self.harmonic_transform(amp[self.power_distributor] * x)
        if self.grid_subslice is None:
            return f
        return f[self.grid_subslice]

    def sandwich(self, x, harmonic_kernel=None) -> NDArray:
        if self.grid_subslice is None:
            x_wpad = x
        else:
            x_wpad = jnp.zeros(tuple(self.grid_shape))
            x_wpad = x_wpad.at[self.grid_subslice].set(x)

        swd = jax.ShapeDtypeStruct(tuple(self.grid_shape), x.dtype)
        ht = self.harmonic_transform
        ht_T = jax.linear_transpose(self.harmonic_transform, swd)

        power = self.power(harmonic_kernel=harmonic_kernel)
        s = ht(power[self.power_distributor] * ht_T(x_wpad)[0])
        if self.grid_subslice is None:
            return s
        return s[self.grid_subslice]

    def __call__(self, x, harmonic_kernel=None) -> NDArray:
        """Applies the Covariance matrix."""
        x_shp = x.shape
        jitter = 0. if self.jitter is None else self.jitter * x

        x = (self.w.T @ x.ravel()).reshape(tuple(self.grid_unpadded_shape))
        x = self.sandwich(x, harmonic_kernel=harmonic_kernel)
        x = (self.w @ x.ravel()).reshape(x_shp)
        return x + jitter

    def evaluate(self, harmonic_kernel=None):
        """Instantiate the full covariance matrix."""
        probe = jnp.zeros(self.w.shape[0])
        indices = jnp.arange(self.w.shape[0]).reshape(1, -1)

        return jax.lax.map(
            lambda idx: self(
                probe.at[tuple(idx)].set(1.), harmonic_kernel=harmonic_kernel
            ).ravel(), indices.T
        ).T  # vmap over `indices` w/ `in_axes=1, out_axes=-1`

    def evaluate_(self, kernel) -> NDArray:
        from scipy.spatial import distance_matrix

        if self.jitter is None:
            jitter = 0.
        else:
            jitter = self.jitter * jnp.eye(self.w.shape[0])

        p = [
            np.linspace(*b, num=sz, endpoint=True) for b, sz in
            zip(self.grid_unpadded_bounds, self.grid_unpadded_shape)
        ]
        p = np.stack(np.meshgrid(*p, indexing="ij"),
                     axis=-1).reshape(-1, len(self.grid_unpadded_shape))
        kernel_inducing = kernel(distance_matrix(p, p))

        return self.w @ kernel_inducing @ self.w.T + jitter


class ToeplitzSKI():
    def __init__(
        self,
        grid_shape: Tuple[int],
        grid_bounds: Tuple[Tuple[float, float]],
        sampling_points: NDArray,
        kernel: Optional[Callable] = None,
        jitter: Union[bool, float, None] = True
    ):
        """Instantiate a KISS-GP model of the covariance using a Toeplitz matrix
        representation of the kernel.

        Parameters
        ----------
        grid_shape :
            Number of pixels along each axes of the inducing points within
            `grid_bounds`.
        grid_bounds :
            Tuple of boundaries of length of the number of dimensions. The
            boundaries should denote the leftmost and rightmost edge of the
            modeling space.
        sampling_points :
            Locations of the modeled points within the grid.
        kernel :
            Position space kernel.
        jitter :
            Strength of the diagonal jitter which to add to the covariance.
        """
        if jitter is True:
            if sampling_points.dtype.type == np.float64:
                self.jitter = 1e-8
            elif sampling_points.dtype.type == np.float32:
                self.jitter = 1e-6
            else:
                raise NotImplementedError()
        elif jitter is False:
            self.jitter = None
        else:
            self.jitter = jitter

        self.grid_shape = np.asarray(grid_shape)
        self.grid_bounds = np.asarray(grid_bounds)
        self.grid_distances = jnp.diff(self.grid_bounds,
                                       axis=1).ravel() / self.grid_shape
        self.ndim = len(grid_shape)

        d = jnp.mgrid[tuple(slice(s) for s in grid_shape)]
        d *= self.grid_distances.reshape((-1, ) + (1, ) * self.ndim)
        self.grid_distances_to_zero = jnp.linalg.norm(d, axis=0)

        self.w = interp_mat(grid_shape, grid_bounds, sampling_points)

        self._kernel = kernel

    @property
    def kernel(self) -> Callable:
        """Yields the harmonic kernel specified during initialization or throw
        a `TypeError`.
        """
        if self._kernel is None:
            te = (
                "either specify a fixed kernel during initialization of the"
                f" {self.__class__.__name__} class or provide one here"
            )
            raise TypeError(te)
        return self._kernel

    def __call__(self, x, kernel=None) -> NDArray:
        """Applies the Covariance matrix."""
        kernel = self.kernel if kernel is None else kernel
        x_shp = x.shape
        jitter = 0. if self.jitter is None else self.jitter * x

        x = (self.w.T @ x.ravel()).reshape(tuple(self.grid_shape))
        cov_row = kernel(self.grid_distances_to_zero)
        x = matmul_toeplitz(cov_row, x)
        x = (self.w @ x.ravel()).reshape(x_shp)
        return x + jitter

    def evaluate(self, kernel=None):
        """Instantiate the full covariance matrix."""
        probe = jnp.zeros(self.w.shape[0])
        indices = jnp.arange(self.w.shape[0]).reshape(1, -1)

        return jax.lax.map(
            lambda idx: self(probe.at[tuple(idx)].set(1.), kernel=kernel).ravel(
            ), indices.T
        ).T  # vmap over `indices` w/ `in_axes=1, out_axes=-1`

    def evaluate_(self, kernel=None) -> NDArray:
        from scipy.spatial import distance_matrix

        kernel = self.kernel if kernel is None else kernel

        if self.jitter is None:
            jitter = 0.
        else:
            jitter = self.jitter * jnp.eye(self.w.shape[0])

        p = [
            np.linspace(*b, num=sz, endpoint=True)
            for b, sz in zip(self.grid_bounds, self.grid_shape)
        ]
        p = np.stack(np.meshgrid(*p, indexing="ij"),
                     axis=-1).reshape(-1, len(self.grid_shape))
        kernel_inducing = kernel(distance_matrix(p, p))

        return self.w @ kernel_inducing @ self.w.T + jitter



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import logging

logger = logging.getLogger(__name__)
c_handler = logging.StreamHandler()
logger.addHandler(c_handler)
logger.setLevel(logging.DEBUG)
logger.propagate = False



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-2-Clause
# Authors: Philipp Frank,

from functools import partial
from typing import Callable, Union

import jax.numpy as jnp
import numpy as np
from jax import Array, vmap
from jax.tree_util import tree_map
from jax.lax import fori_loop

from .model import Initializer, LazyModel, Model
from .prior import LogNormalPrior, NormalPrior
from .tree_math import ShapeWithDtype, random_like


def _isscalar(x):
    return jnp.ndim(x) == 0


def discrete_gauss_markov_process(
    xi: Array, x0: Array, drift: Array, diffamp: Array
):
    """Generator for a Gauss-Markov process  (GMP).

    Given the discrete transition probabilities via the `drift` and `diffamp`
    matrices, this function produces a series `res` of the form:

    .. math::
        res_{i+1} = drift_i @ res_i + diffamp_i @ xi_i

    which corresponds to a random realization of a GMP in case `xi` is
    standard normal distributed.

    Parameters:
    -----------
    xi: Array
        Input random variables for the GMP.
    x0: Array
        Initial state of the process. The first entry of the result `res_0` is
        always equal to `x0`.
    drift: Array
        Matrix or sequence of matrices that describe the mean of the transition
        probabilities of the GMP.
    diffamp: Array
        Matrix or sequence of matrices that correspond to the amplitude of the
        transition probabilities. For the GMP they imply that their matrix
        product `diffamp_i @ diffamp_i.T` is equal to the covariance of the
        transition probability iff `xi` is standard normal.

    Returns
    -------
        Sequence of vectors of the Gauss-Markov series. The sequence is returned
        as an array where the first axis corresponds to the steps of the
        sequence and the second axis are the entries of the vector.
    Notes:
    ------
    In case the sequence `xi` has length N (i.E. `xi.shape[0] = N`) the result
    sequence is of length N+1 as `x0` is returned as the first entry of the
    final sequence, followed by N transitions.
    """
    if _isscalar(drift):
        drift = drift * jnp.ones((1, 1))
    if _isscalar(diffamp):
        diffamp = diffamp * jnp.ones((1, 1))

    in_ax = (None if len(diffamp.shape) == 2 else 0, 0)
    res = vmap(jnp.matmul, in_ax, 0)(diffamp, xi)

    def loop(i, a):
        d = drift[i] if len(drift.shape) > 2 else drift
        return a.at[i + 1].add(jnp.matmul(d, a[i]))

    res = jnp.concatenate([x0[jnp.newaxis, ...], res], axis=0)
    return fori_loop(0, res.size, loop, res)


def scalar_gauss_markov_process(xi, x0, drift, diffamp):
    """Simple wrapper of `discrete_gm_general` for 1D scalar processes.
    """
    if not _isscalar(drift):
        drift = drift[:, jnp.newaxis, jnp.newaxis]
    if not _isscalar(diffamp):
        diffamp = diffamp[:, jnp.newaxis, jnp.newaxis]
    if _isscalar(x0):
        x0 = jnp.array([x0])
    return discrete_gauss_markov_process(
        xi[:, jnp.newaxis], x0, drift, diffamp
    )[:, 0]


def wiener_process(
    xi: Array,
    x0: Union[float, Array],
    sigma: Union[float, Array],
    dt: Union[float, Array],
):
    """Implements the Wiener process (WP)."""
    amp = jnp.sqrt(dt) * sigma
    return jnp.cumsum(jnp.concatenate((jnp.atleast_1d(x0).flatten(), amp * xi)))


def integrated_wiener_process(
    xi: Array,
    x0: Array,
    sigma: Array,
    dt: Array,
    asperity: Union[float, Array] = None
):
    """Implements the (generalized) Integrated Wiener process (IWP)."""
    asperity = 0. if asperity is None else asperity
    dt = jnp.ones(xi.shape[0]) * dt if _isscalar(dt) else dt
    res = (sigma * jnp.sqrt(dt))[:, jnp.newaxis] * xi
    res = res.at[:, 0].mul(jnp.sqrt(dt**2 / 12. + asperity))
    res = res.at[:, 0].add(0.5 * dt * res[:, 1])
    res = jnp.concatenate((x0[jnp.newaxis, ...], res), axis=0)
    res = res.at[:, 1].set(jnp.cumsum(res[:, 1]))
    res = res.at[1:, 0].add(dt * res[:-1, 1])
    return res.at[:, 0].set(jnp.cumsum(res[:, 0]))


def ornstein_uhlenbeck_process(
    xi: Array, x0: float, sigma: Union[float, Array],
    gamma: Union[float, Array], dt: Union[float, Array]
):
    """Implements the Ornstein Uhlenbeck process (OUP)."""
    drift = jnp.exp(-gamma * dt)
    amp = sigma * jnp.sqrt(1. - drift**2)
    return scalar_gauss_markov_process(xi, x0, drift, amp)


class GaussMarkovProcess(Model):
    def __init__(
        self,
        process: Callable,
        x0: Union[float, Array, LazyModel],
        dt: Union[float, Array],
        name='xi',
        N_steps: int = None,
        **kwargs
    ):
        if _isscalar(dt):
            if N_steps is None:
                msg = "`N_steps` is None and `dt` is not a sequence"
                raise NotImplementedError(msg)
            dt = np.ones(N_steps) * dt
        shp = dt.shape + jnp.shape(
            x0.target if isinstance(x0, LazyModel) else x0
        )
        domain = {name: ShapeWithDtype(shp)}
        init = Initializer(
            tree_map(lambda x: partial(random_like, primals=x), domain)
        )
        if isinstance(x0, LazyModel):
            domain = domain | x0.domain
            init = init | x0.init
        self.x0 = x0
        for _, a in kwargs.items():
            if isinstance(a, LazyModel):
                domain = domain | a.domain
                init = init | a.init
        self.kwargs = kwargs
        self.name = name
        self.process = process
        self.dt = dt

        super().__init__(domain=domain, init=init)

    def __call__(self, x):
        xi = x[self.name]
        xx = self.x0(x) if isinstance(self.x0, LazyModel) else self.x0
        tmp = {
            k: a(x) if isinstance(a, LazyModel) else a
            for k, a in self.kwargs.items()
        }
        return self.process(xi=xi, x0=xx, dt=self.dt, **tmp)


def WienerProcess(
    x0: Union[tuple, float, LazyModel],
    sigma: Union[tuple, float, Array, LazyModel],
    dt: Union[float, Array],
    name: str = 'wp',
    N_steps: int = None
):
    """Implements the Wiener process (WP).

    The WP in continuous time takes the form:

    .. math::
        d/dt x_t = sigma xi_t ,

    where `xi_t` is continuous time white noise.

    Parameters:
    -----------
    x0: tuple, float, or LazyModel
        Initial position of the WP. Can be passed as a fixed value, or a
        generative Model. Passing a tuple is a shortcut to set a normal prior
        with mean and std equal to the first and second entry of the tuple
        respectively on `x0`.
    sigma: tuple, float, Array, LazyModel
        Standard deviation of the WP. Analogously to `x0` may also be passed on
        as a model. May also be passed as a sequence of length equal to `dt` in
        which case a different sigma is used for each time interval.
    dt: float or Array of float
        Stepsizes of the process. In case it is a single float, `N_steps` must
        be provided to indicate the number of steps taken.
    name: str
        Name of the key corresponding to the parameters of the WP. Default `wp`.
    N_steps: int (optional)
        Option to set the number of steps in case `dt` is a scalar.

    Notes:
    ------
    In case `sigma` is time dependent, i.E. passed on as a sequence
    of length equal to `xi`, it is assumed to be constant within each timebin,
    i.E. `sigma_t = sigma_i for t_i <= t < t_{i+1}`.
    """
    if isinstance(x0, tuple):
        x0 = NormalPrior(x0[0], x0[1], name=name + '_x0')
    if isinstance(sigma, tuple):
        sigma = LogNormalPrior(sigma[0], sigma[1], name=name + '_sigma')
    return GaussMarkovProcess(
        wiener_process, x0, dt, name=name, N_steps=N_steps, sigma=sigma
    )


def IntegratedWienerProcess(
    x0: Union[tuple, Array, LazyModel],
    sigma: Union[tuple, float, Array, LazyModel],
    dt: Union[float, Array],
    name: str = 'iwp',
    asperity: Union[tuple, float, Array, LazyModel] = None,
    N_steps: int = None
):
    """Implements the Integrated Wiener Process.

    The generalized IWP in continuous time takes the form:

    .. math::
        d/dt x_t = y_t + sigma * asperity xi^1_t , \\
        d/dt y_t = sigma * xi^2_t

    where `xi^i_t` are continuous time white noise processes.
    This is a standard IWP in x in the case that `asperity` is zero (None). If
    it is non-zero, a WP with relative strength set by `asperity` is added to
    the dynamics of the IWP.

    For information on the parameters, please refer to `WienerProcess`. The
    `asperity` parameter may be defined analogously to `sigma`.

    This is also the process used in the CorrelatedField to describe the
    deviations of the power spectra from power-laws on a double logarithmic
    scale. See also `re.correlated_field.CorrelatedFieldMaker` and references
    there for further information.

    Notes:
    ------
    `sigma` and `asperity` may also be sequences. See notes on `WienerProcess`
    for further information.
    """
    if isinstance(x0, tuple):
        x0 = NormalPrior(x0[0], x0[1], shape=(2, ), name=name + '_x0')
    if isinstance(sigma, tuple):
        sigma = LogNormalPrior(sigma[0], sigma[1], name=name + '_sigma')
    if isinstance(asperity, tuple):
        asperity = LogNormalPrior(
            asperity[0], asperity[1], name=name + '_asperity'
        )
    return GaussMarkovProcess(
        integrated_wiener_process,
        x0,
        dt,
        name=name,
        N_steps=N_steps,
        sigma=sigma,
        asperity=asperity
    )


def OrnsteinUhlenbeckProcess(
    sigma: Union[tuple, float, Array, LazyModel],
    gamma: Union[tuple, float, Array, LazyModel],
    dt: Union[float, Array],
    name: str = 'oup',
    x0: Union[tuple, float, LazyModel] = None,
    N_steps: int = None
):
    """Implements the Ornstein Uhlenbeck process (OUP).

    The stochastic differential equation of the OUP takes the form:

    .. math::
        d/dt x_t + gamma x_t = sigma xi_t

    where `xi_t` is continuous time white noise.

    For information on the parameters, please refer to `WienerProcess`. The
    `gamma` parameter may be defined analogously to `sigma`.
    Unlike the WP and IWP, the OUP is a proper process, i.E. allows for a proper
    steady state solution. Therefore, `x0` may not be provided at all, in which
    case it defaults to a random variable drawn from the steady state
    distribution of the OUP.

    Notes:
    ------
    `sigma` and `gamma` may also be sequences. See notes on `WienerProcess`
    for further information.
    """
    if isinstance(sigma, tuple):
        sigma = LogNormalPrior(sigma[0], sigma[1], name=name + '_sigma')
    if isinstance(gamma, tuple):
        gamma = LogNormalPrior(gamma[0], gamma[1], name=name + '_gamma')
    if x0 is None:
        # x0 is not set, use steady state distribution of OUP to generate it.
        key = name + '_x0'

        def gen_x0(x):
            res = x[key]
            sig = sigma(x) if isinstance(sigma, LazyModel) else sigma
            return res * (sig if _isscalar(sig) else sig[0])

        domain = {key: ShapeWithDtype(())}
        init = Initializer(
            tree_map(lambda x: partial(random_like, primals=x), domain)
        )
        if isinstance(sigma, LazyModel):
            domain = domain | sigma.domain
            init = init | sigma.init
        x0 = Model(gen_x0, domain=domain, init=init)
    elif isinstance(x0, tuple):
        x0 = NormalPrior(x0[0], x0[1], name=name + '_x0')
    return GaussMarkovProcess(
        ornstein_uhlenbeck_process,
        x0,
        dt,
        name=name,
        N_steps=N_steps,
        sigma=sigma,
        gamma=gamma
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import nifty8 as ift
import numpy as np
from numpy.testing import assert_allclose

from ..common import setup_function, teardown_function


def test_conjugation_operator():
    sp = ift.RGSpace(8)
    dom = ift.makeDomain(sp)
    f = ift.from_random(dom, dtype=np.complex128)
    op = ift.ScalingOperator(sp, 1).conjugate()
    arr = f.val
    res1 = f.conjugate()
    res2 = op(f)
    res3 = arr.conjugate()
    assert_allclose(res1.val, res2.val)
    assert_allclose(res1.val, res3)
    ift.extra.check_linear_operator(op, domain_dtype=np.float64,
                                    target_dtype=np.float64)
    ift.extra.check_linear_operator(op, domain_dtype=np.complex128,
                                    target_dtype=np.complex128, only_r_linear=True)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
# Author: Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from warnings import warn

import numpy as np

from ..domain_tuple import DomainTuple
from ..domains.power_space import PowerSpace
from ..operators.adder import Adder
from ..operators.contraction_operator import ContractionOperator
from ..operators.distributors import PowerDistributor
from ..operators.harmonic_operators import HarmonicTransformOperator
from ..operators.normal_operators import LognormalTransform, NormalTransform
from ..operators.simple_linear_operators import ducktape
from ..operators.value_inserter import ValueInserter
from ..sugar import full, makeField, makeOp
from .correlated_fields import (_log_vol, _Normalization,
                                _relative_log_k_lengths, _SlopeRemover,
                                _TwoLogIntegrations)


def SimpleCorrelatedField(
    target,
    offset_mean,
    offset_std,
    fluctuations,
    flexibility,
    asperity,
    loglogavgslope,
    prefix="",
    harmonic_partner=None,
):
    """Simplified version of :class:`~nifty8.library.correlated_fields.CorrelatedFieldMaker`.

    Assumes `total_N = 0`, `dofdex = None` and the presence of only one power
    spectrum, i.e. only one call of
    :func:`~nifty8.library.correlated_fields.CorrelatedFieldMaker.add_fluctuations`.

    See also
    --------
    * The simple correlated field model has first been described in "Comparison
      of classical and Bayesian imaging in radio interferometry", A&A 646, A84
      (2021) by P. Arras et al.
      `<https://doi.org/10.1051/0004-6361/202039258>`_

    Consider citing this paper, if you use the simple correlated field model.
    """
    target = DomainTuple.make(target)
    if len(target) != 1:
        raise ValueError
    target = target[0]
    if harmonic_partner is None:
        harmonic_partner = target.get_default_codomain()
    else:
        target.check_codomain(harmonic_partner)
        harmonic_partner.check_codomain(target)
    for kk in (fluctuations, loglogavgslope):
        if len(kk) != 2:
            raise TypeError
    for kk in (offset_std, flexibility, asperity):
        if not (kk is None or len(kk) == 2):
            raise TypeError
    if flexibility is None and asperity is not None:
        raise ValueError
    fluct = LognormalTransform(*fluctuations, prefix + 'fluctuations', 0)
    avgsl = NormalTransform(*loglogavgslope, prefix + 'loglogavgslope', 0)

    pspace = PowerSpace(harmonic_partner)
    twolog = _TwoLogIntegrations(pspace)
    expander = ContractionOperator(twolog.domain, 0).adjoint
    ps_expander = ContractionOperator(pspace, 0).adjoint
    vslope = makeOp(makeField(pspace, _relative_log_k_lengths(pspace)))
    slope = vslope @ ps_expander @ avgsl
    a = slope

    if flexibility is not None:
        flex = LognormalTransform(*flexibility, prefix + 'flexibility', 0)
        dom = twolog.domain[0]
        vflex = np.empty(dom.shape)
        vflex[0] = vflex[1] = np.sqrt(_log_vol(pspace))
        vflex = makeOp(makeField(dom, vflex))
        sig_flex = vflex @ expander @ flex
        xi = ducktape(dom, None, prefix + 'spectrum')

        shift = np.empty(dom.shape)
        shift[0] = _log_vol(pspace)**2 / 12.
        shift[1] = 1
        shift = makeField(dom, shift)
        if asperity is None:
            asp = makeOp(shift.ptw("sqrt")) @ (xi*sig_flex)
        else:
            asp = LognormalTransform(*asperity, prefix + 'asperity', 0)
            vasp = np.empty(dom.shape)
            vasp[0] = 1
            vasp[1] = 0
            vasp = makeOp(makeField(dom, vasp))
            sig_asp = vasp @ expander @ asp
            asp = xi*sig_flex*(Adder(shift) @ sig_asp).ptw("sqrt")
        a = a + _SlopeRemover(pspace, 0) @ twolog @ asp
    a = _Normalization(pspace, 0) @ a
    maskzm = np.ones(pspace.shape)
    maskzm[0] = 0
    maskzm = makeOp(makeField(pspace, maskzm))
    a = (maskzm @ ((ps_expander @ fluct)*a))
    if offset_std is not None:
        zm = LognormalTransform(*offset_std, prefix + 'zeromode', 0)
        insert = ValueInserter(pspace, (0,))
        a = a + insert(zm)
    a = a.scale(target.total_volume)

    ht = HarmonicTransformOperator(harmonic_partner, target)
    pd = PowerDistributor(harmonic_partner, pspace)
    xi = ducktape(harmonic_partner, None, prefix + 'xi')
    op = ht(pd(a)*xi)
    if offset_mean is not None:
        op = Adder(full(op.target, float(offset_mean))) @ op
    op.amplitude = a
    op.power_spectrum = a**2

    try:
        from .. import RGSpace
        from .. import re as jft

        if not all(isinstance(dom, RGSpace) for dom in op.target):
            # warn(f"unable to add JAX operator for {op.target!r}")
            raise ImportError("short-circuit JAX init")

        dists = tuple(e for di in op.target for e in di.distances)
        cfm = jft.CorrelatedFieldMaker(prefix=prefix)
        cfm.add_fluctuations(
            shape=op.target.shape,
            distances=dists,
            fluctuations=fluctuations,
            loglogavgslope=loglogavgslope,
            flexibility=flexibility,
            asperity=asperity,
            prefix="",
            harmonic_type="fourier",
            non_parametric_kind="power",
        )
        cfm.set_amplitude_total_offset(
            offset_mean=offset_mean, offset_std=offset_std
        )
        cf = cfm.finalize()

        op._jax_expr = cf
        op.amplitude._jax_expr = cfm.amplitude
        op.power_spectrum._jax_expr = cfm.power_spectrum
    except (ImportError, TypeError) as e:
        pass
        # if isinstance(e, TypeError):
        #     warn(f"no JAX operator for this configuration;\n{e}")

    return op



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .operators.operator import Operator
from .sugar import makeOp
from .utilities import check_object_identity


class Linearization(Operator):
    """Let `A` be an operator and `x` a field. `Linearization` stores the value
    of the operator application (i.e. `A(x)`), the local Jacobian
    (i.e. `dA(x)/dx`) and, optionally, the local metric.

    Parameters
    ----------
    val : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
        The value of the operator application.
    jac : LinearOperator
        The Jacobian.
    metric : LinearOperator or None
        The metric. Default: None.
    want_metric : bool
        If True, the metric will be computed for other Linearizations derived
        from this one. Default: False.
    """
    def __init__(self, val, jac, metric=None, want_metric=False):
        self._val = val
        self._jac = jac
        check_object_identity(self._val.domain, self._jac.target)
        self._want_metric = want_metric
        self._metric = metric

    def new(self, val, jac, metric=None):
        """Create a new Linearization, taking the `want_metric` property from
           this one.

        Parameters
        ----------
        val : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            the value of the operator application
        jac : LinearOperator
            the Jacobian
        metric : LinearOperator or None
            The metric. Default: None.
        """
        return Linearization(val, jac, metric, self._want_metric)

    def trivial_jac(self):
        return self.make_var(self._val, self._want_metric)

    def prepend_jac(self, jac):
        if self._metric is None:
            return self.new(self._val, self._jac @ jac)
        from .operators.sandwich_operator import SandwichOperator
        metric = SandwichOperator.make(jac, self._metric)
        return self.new(self._val, self._jac @ jac, metric)

    @property
    def domain(self):
        """DomainTuple or MultiDomain : the Jacobian's domain"""
        return self._jac.domain

    @property
    def target(self):
        """DomainTuple or MultiDomain : the Jacobian's target (i.e. the value's domain)"""
        return self._jac.target

    @property
    def val(self):
        """:class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField` : the value"""
        return self._val

    @property
    def jac(self):
        """LinearOperator : the Jacobian"""
        return self._jac

    @property
    def gradient(self):
        """:class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField` : the gradient

        Notes
        -----
        Only available if target is a scalar
        """
        from .field import Field
        return self._jac.adjoint_times(Field.scalar(1.))

    @property
    def want_metric(self):
        """bool : True iff the metric was requested in the constructor"""
        return self._want_metric

    @property
    def metric(self):
        """LinearOperator : the metric

        Notes
        -----
        Only available if target is a scalar
        """
        return self._metric

    def __getitem__(self, name):
        return self.new(self._val[name], self._jac.ducktape_left(name))

    def __neg__(self):
        if self._metric is not None:
            raise RuntimeError("Cannot negate operators with metric")
        return self.new(-self._val, -self._jac, metric=None)

    def conjugate(self):
        return self.new(
            self._val.conjugate(), self._jac.conjugate(),
            None if self._metric is None else self._metric.conjugate())

    @property
    def real(self):
        return self.new(self._val.real, self._jac.real)

    @property
    def imag(self):
        return self.new(self._val.imag, self._jac.imag)

    def _myadd(self, other, neg):
        if np.isscalar(other) or other.jac is None:
            return self.new(self._val-other if neg else self._val+other,
                            self._jac, self._metric)
        met = None
        if self._metric is not None and other._metric is not None:
            met = self._metric._myadd(other._metric, neg)
        return self.new(
            self.val.flexible_addsub(other.val, neg),
            self.jac._myadd(other.jac, neg), met)

    def __add__(self, other):
        return self._myadd(other, False)

    def __radd__(self, other):
        return self._myadd(other, False)

    def __sub__(self, other):
        return self._myadd(other, True)

    def __rsub__(self, other):
        return (-self).__add__(other)

    def __truediv__(self, other):
        if np.isscalar(other):
            return self.__mul__(1/other)
        return self.__mul__(other.ptw("reciprocal"))

    def __rtruediv__(self, other):
        return self.ptw("reciprocal").__mul__(other)

    def __pow__(self, power):
        if not (np.isscalar(power) or power.jac is None):
            return NotImplemented
        return self.ptw("power", power)

    def __mul__(self, other):
        if np.isscalar(other):
            if other == 1:
                return self
            met = None if self._metric is None else self._metric.scale(other)
            return self.new(self._val*other, self._jac.scale(other), met)
        if other.jac is None:
            check_object_identity(self.target, other.domain)
            return self.new(self._val*other, makeOp(other)(self._jac))
        check_object_identity(self.target, other.target)
        return self.new(
            self.val*other.val,
            (makeOp(other.val)(self.jac))._myadd(
             makeOp(self.val)(other.jac), False))

    def __rmul__(self, other):
        return self.__mul__(other)

    def outer(self, other):
        """Computes the outer product of this Linearization with a Field or
        another Linearization

        Parameters
        ----------
        other : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField` or Linearization

        Returns
        -------
        Linearization
            the outer product of self and other
        """
        if np.isscalar(other):
            return self.__mul__(other)
        from .operators.outer_product_operator import OuterProduct
        if other.jac is None:
            return self.new(OuterProduct(other.domain, self._val)(other),
                            OuterProduct(other.domain, self._jac(self._val)))
        tmp_op = OuterProduct(other.target, self._val)
        return self.new(
            tmp_op(other._val),
            OuterProduct(other.target, self._jac(self._val))._myadd(
                tmp_op(other._jac), False))

    def vdot(self, other):
        """Computes the inner product of this Linearization with a Field or
        another Linearization

        Parameters
        ----------
        other : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField` or Linearization

        Returns
        -------
        Linearization
            the inner product of self and other
        """
        from .operators.simple_linear_operators import VdotOperator
        if other.jac is None:
            return self.new(
                self._val.vdot(other),
                VdotOperator(other)(self._jac))
        return self.new(
            self._val.vdot(other._val),
            VdotOperator(self._val)(other._jac) +
            VdotOperator(other._val)(self._jac))

    def sum(self, spaces=None):
        """Computes the (partial) sum over self

        Parameters
        ----------
        spaces : None, int or list of int
            - if None, sum over the entire domain
            - else sum over the specified subspaces

        Returns
        -------
        Linearization
            the (partial) sum
        """
        from .operators.contraction_operator import ContractionOperator
        return self.new(
            self._val.sum(spaces),
            ContractionOperator(self._jac.target, spaces)(self._jac))

    def integrate(self, spaces=None):
        """Computes the (partial) integral over self

        Parameters
        ----------
        spaces : None, int or list of int
            - if None, integrate over the entire domain
            - else integrate over the specified subspaces

        Returns
        -------
        Linearization
            the (partial) integral
        """
        from .operators.contraction_operator import IntegrationOperator
        return IntegrationOperator(self.target, spaces)(self)

    def ptw(self, op, *args, **kwargs):
        t1, t2 = self._val.ptw_with_deriv(op, *args, **kwargs)
        return self.new(t1, makeOp(t2)(self._jac))

    def add_metric(self, metric):
        return self.new(self._val, self._jac, metric)

    def with_want_metric(self):
        return Linearization(self._val, self._jac, self._metric, True)

    @staticmethod
    def make_var(field, want_metric=False):
        """Converts a Field to a Linearization, with a unity Jacobian

        Parameters
        ----------
        field : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            the field to be converted
        want_metric : bool
            If True, the metric will be computed for other Linearizations
            derived from this one. Default: False.

        Returns
        -------
        Linearization
            the requested Linearization
        """
        from .operators.scaling_operator import ScalingOperator
        return Linearization(field, ScalingOperator(field.domain, 1.),
                             want_metric=want_metric)

    @staticmethod
    def make_const(field, want_metric=False):
        """Converts a Field to a Linearization, with a zero Jacobian

        Parameters
        ----------
        field : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            the field to be converted
        want_metric : bool
            If True, the metric will be computed for other Linearizations
            derived from this one. Default: False.

        Returns
        -------
        Linearization
            the requested Linearization

        Notes
        -----
        The Jacobian is square and contains only zeroes.
        """
        from .operators.simple_linear_operators import NullOperator
        return Linearization(field, NullOperator(field.domain, field.domain),
                             want_metric=want_metric)

    @staticmethod
    def make_const_empty_input(field, want_metric=False):
        """Converts a Field to a Linearization, with a zero Jacobian

        Parameters
        ----------
        field : :class:`nifty8.field.Field` or :class:`nifty8.multi_field.MultiField`
            the field to be converted
        want_metric : bool
            If True, the metric will be computed for other Linearizations
            derived from this one. Default: False.

        Returns
        -------
        Linearization
            the requested Linearization

        Notes
        -----
        The Jacobian has an empty input domain, i.e. its matrix representation
        has 0 columns.
        """
        from .multi_domain import MultiDomain
        from .operators.simple_linear_operators import NullOperator
        return Linearization(
            field, NullOperator(MultiDomain.make({}), field.domain),
            want_metric=want_metric)

    @staticmethod
    def make_partial_var(field, constants, want_metric=False):
        """Converts a MultiField to a Linearization, with a Jacobian that is
        unity for some MultiField components and a zero matrix for others.

        Parameters
        ----------
        field ::class:`nifty8.multi_field.MultiField`
            the field to be converted
        constants : list of string
            the MultiField components for which the Jacobian should be
            a zero matrix.
        want_metric : bool
            If True, the metric will be computed for other Linearizations
            derived from this one. Default: False.

        Returns
        -------
        Linearization
            the requested Linearization

        Notes
        -----
        The Jacobian is square.
        """
        from .operators.block_diagonal_operator import BlockDiagonalOperator
        from .operators.scaling_operator import ScalingOperator
        if len(constants) == 0:
            return Linearization.make_var(field, want_metric)
        else:
            ops = {key: ScalingOperator(dom, 0. if key in constants else 1.)
                   for key, dom in field.domain.items()}
            bdop = BlockDiagonalOperator(field.domain, ops)
            return Linearization(field, bdop, want_metric=want_metric)



import nifty8

needs_sphinx = '3.2.0'

extensions = [
    'sphinx.ext.napoleon',    # Support for NumPy and Google style docstrings
    'sphinx.ext.imgmath',     # Render math as images
    'sphinx.ext.viewcode',    # Add links to highlighted source code
    'sphinx.ext.intersphinx', # Links to other sphinx docs (mostly numpy)
    'myst_parser',            # Parse markdown
]
master_doc = 'index'

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
    "strikethrough",
    "tasklist",
]

intersphinx_mapping = {"numpy": ("https://numpy.org/doc/stable/", None),
                       #"matplotlib": ('https://matplotlib.org/stable/', None),
                       "ducc0": ("https://mtr.pages.mpcdf.de/ducc/", None),
                       "scipy": ('https://docs.scipy.org/doc/scipy/reference/', None),
                       }

autodoc_default_options = {'special-members': '__init__'}

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_ivar = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_references = True
napoleon_include_special_with_doc = True

imgmath_embed = True

project = u'NIFTy8'
copyright = u'2013-2022, Max-Planck-Society'
author = u'Martin Reinecke'

release = nifty8.version.__version__
version = release[:-2]

language = "en"
exclude_patterns = []
add_module_names = False

html_theme = "pydata_sphinx_theme"
html_context = {"default_mode": "light"}
html_logo = 'nifty_logo_black.png'

html_theme_options = {
    "logo": {
        "image_light": "nifty_logo_black.png",
        "image_dark": "nifty_logo_black.png"
    },
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/nifty8",
            "icon": "fas fa-box",
        },
        {
            "name": "GitLab",
            "url": "https://gitlab.mpcdf.mpg.de/ift/nifty",
            "icon": "fab fa-gitlab",
        }
    ],
    "navbar_persistent": ["search-field"],
    "navbar_end": ["navbar-icon-links"],
}

html_last_updated_fmt = '%b %d, %Y'

exclude_patterns = ['mod/modules.rst', 'mod/nifty8.logger.rst']
